/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: lineio.c,v 1.6 2000/12/20 15:45:00 bazsi Exp $
 *
 * Author  : Buci
 * Auditor : kisza
 * Last version: 1.2
 * Notes:
 *    - GNU indenting
 *   - comments, in English
 *   - GNU indenting
 *   - read_line(): cnt grows, grows, withouth any checking, and
 *        used for indexing ( buff[], which has a fix length )
 *   - write_token_line(): gcnt not checked, but used for string indexing
 *
 ***************************************************************************/

#include <zorp/lineio.h>

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

void free_g_list_of_strings(GList * list);

/*
   Reads one line from the given GIOChannel, and cuts it into tokens.
   A line ends with a '\n' character, and tokens are delimited by
   exactly one space. A token may contain space or '\n' escaped
   by a backslash.
*/

GList *
read_token_line(GIOChannel * channel)
{
  GString *str = NULL;

  GList *list = NULL;
  char *element;
  gchar a;
  int ret, protect = FALSE, bytes_read;

  g_return_val_if_fail(channel, NULL);

  str = g_string_new("");

  while (1)
    {
      ret = g_io_channel_read(channel, &a, 1, &bytes_read);
      if (ret != G_IO_ERROR_NONE)
	{
	  free_g_list_of_strings(list);
	  return NULL;
	}
      if (bytes_read != 1)
	a = '\n';
      if ((a == '\n' || a == ' ') && !protect)
	{
	  element = g_new0(char, str->len + 1);
	  strncpy(element, str->str, str->len);
	  list = g_list_append(list, element);
	  g_string_assign(str, "");
	}
      else if (a == '\\' && !protect)
	{
	  protect = 1;
	}
      else
	{
	  g_string_append_c(str, a);
	  protect = 0;
	}
      if (a == '\n')
	break;
    }
  return list;
}

/*
 *  'frontend' for g_free accepting two parameters, but the secount is
 *  discarded. Needed when useing g_list_foreach.
 */
static void
b_free(gpointer str, gpointer a)
{
  g_free(str);
}

/*
 *  Frees a list created by the read_token_line function.
 *  (or any other list, that has elements which are pointers.)
 *  It frees all the elements as well.
 */
void
free_g_list_of_strings(GList * list)
{
  g_list_foreach(list, b_free, NULL);
  g_list_free(list);
}

/*
 *  Reads in one line from 'channel' which contains tokens
 *  (as defined before the read_token_line function), and checks if
 *  it is a command that is contained in the 'cmds' array.
 *  If yes, '*which' equals the the index of the command in the 'cmds'
 *  array, and '*parnum' the number of paramerters.
 *  Returns the token liston success, or NULL on error.
 */
GList *
read_command(GIOChannel * channel, Command * cmds, int *which, int *parnum)
{
  GList *line, *p;
  int cnt, i;

  line = read_token_line(channel);
  if (!line)
    {
      *which = -1;
      *parnum = 0;
      return NULL;
    }

  cnt = 0;
  p = line;
  while (p)
    {
      cnt++;
      p = p->next;
    }
  cnt--;

  if (cnt < 0)
    {				// bar ha jobban vegiggondoljuk, ez lehetetlen
      *which = -1;
      *parnum = 0;
      free_g_list_of_strings(line);
      return NULL;
    }

  i = 0;
  while (cmds[i].cmd)
    {
      if ((cnt == cmds[i].argnum || cmds[i].argnum == -1) &&
	  !strcmp(cmds[i].cmd, line->data))
	{
	  *which = i;
	  *parnum = cnt;
	  return line;
	}
      i++;
    }

  *which = -1;
  *parnum = 0;
  free_g_list_of_strings(line);
  return NULL;
}

/*
 *  Returns wether a character needs to be protected. (see definition of a
 *  token line before the read_token_line function)
 */
static int
is_protected(char a)
{
  if (a == '\\')
    return 1;
  if (a == ' ')
    return 1;
  if (a == '\n')
    return 1;
  return 0;
}

/*
 *  appends a atring to 'str', and protects with a backslash each character
 *  that is told to be protected by the is_protected function.
 */
void
g_string_append_protected(GString * gstr, char *str)
{
  int cnt;
  gboolean protect = FALSE;

  cnt = 0;
  while (str[cnt])
    {
      if (!protect && is_protected(str[cnt]))
	{
	  g_string_append_c(gstr, '\\');
	  protect = TRUE;
	}
      else
	{
	  g_string_append_c(gstr, str[cnt++]);
	  protect = FALSE;
	}
    }
}

/*
 *  writes a token line to 'channel'. 'fmt' describes the last paramters.
 *  a character 's' in 'fmt' means a simple string.
 *  a character 'a' in 'fmt' means an array of strings, where the last
 *     value in the array is NULL, and each string is a token
 *  a character 'g' in 'fmt' means a GList, where each element is a pointer
 *     to a string and each string is a token.
 */
int
write_token_line(GIOChannel * channel, char *fmt, ...)
{
  static GString *str = NULL;


  char *current;
  GList *list;
  char **array;

  int i, fcnt;
  int ret, bytes_written;

  va_list args;

  va_start(args, fmt);

  // once in a lifetime we initialise the 'str'. (it is static :)  )
  if (!str)
    str = g_string_new("");

  // but evrytime we have to reset it to ""
  g_string_assign(str, "");

  fcnt = 0;
  while (fmt[fcnt])
    {
      switch (fmt[fcnt])
	{
	case 's':
	  current = va_arg(args, char *);
	  g_string_append_protected(str, current);
	  if (fmt[fcnt + 1])
	    g_string_append_c(str, ' ');
	  else
	    g_string_append_c(str, '\n');
	  break;
	case 'g':
	  list = va_arg(args, GList *);
	  while (list)
	    {
	      current = list->data;
	      g_string_append_protected(str, current);

	      if (list->next)
		g_string_append_c(str, ' ');
	      else if (fmt[fcnt + 1])
		g_string_append_c(str, ' ');
	      else
		g_string_append_c(str, '\n');
	      list = list->next;
	    }
	  break;
	case 'a':
	  i = 0;
	  array = va_arg(args, char **);
	  while (array[i])
	    {
	      current = array[i];
	      g_string_append_protected(str, current);

	      if (array[i + 1])
		g_string_append_c(str, ' ');
	      else if (fmt[fcnt + 1])
		g_string_append_c(str, ' ');
	      else
		g_string_append_c(str, '\n');
	      i++;
	    }
	  break;
	default:
	  break;
	}
      fcnt++;
    }
  va_end(args);

#if 0
#ifdef Z_CURSES
  printw("%s\n", buffer);
#else
  printf("%s\n", buffer);
#endif
#endif

  ret = g_io_channel_write(channel, str->str, str->len, &bytes_written);
  return ret;
}
