/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: packsock.c,v 1.7.2.1 2001/02/22 14:06:27 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.4
 * Notes:
 *    - messages: no process identifier !
 *
 ***************************************************************************/

/* Where is the glib equivalent for realloc()? */

#define MAX_UDP_PACKET 16384
#define MAX_TIMEOUT 15

#include <time.h>

#include <zorp/packsock.h>
#include <zorp/sockaddr.h>
#include <zorp/zorp.h>
#include <zorp/io.h>
#include <zorp/log.h>
#include <zorp/cap.h>

#define CORE_DEBUG "core.debug"
#define CORE_ERROR "core.error"

/*
 * These structures form a linked list in ZPacketSocket's. One of these
 * represents a special from/to/port pair that must be sent to an fd.
 */

typedef struct _ZPacketDispatcher
{
  ZSockAddr *from;
  ZSockAddr *to;
  int fd;
  struct _ZPacketDispatcher *next;
  time_t timeout;
} ZPacketDispatcher;

/*
 * This is the representation of an UDP port. A hash table (z_packsock_table)
 * is constructed from these to avoid storing all 65536 of them in memory
 */

typedef struct _ZPacketSocket
{
  int port;			/* port number */
  int inet_fd;			/* fd of AF_INET socket */
  int default_fd;		/* where to send packets */
  ZPacketDispatcher *special_fd;	/* special packets */
  ZIOWatchFD *watch;
} ZPacketSocket;

/* Ghash means 'fire' in orcish :)   - Tolkien rulez! */

static GHashTable *z_packsock_table = NULL;
static int z_packsock_socket = -1;

/* Keys of the hash table are int's. */

gint
z_packsock_compare(gconstpointer k1, gconstpointer k2)
{
  int p1 = (int) k1;		/* Should use the official conversion func? */
  int p2 = (int) k2;

  return p1 == p2;
}

guint
z_packsock_hash(gconstpointer k)
{
  int p = (int) k;

  return p % 16;		/* hash into sixteen hash buckets */
}

void
z_packsock_init(void)
{
  if (z_packsock_table == NULL)
    {
      z_packsock_table = g_hash_table_new(z_packsock_hash,
					  z_packsock_compare);
    }
}

/*
 * Checks if a socket can be closed (i.e. there are no associated fd's,
 * and closes it if so.
 */

static void
z_packsock_check_destroy(ZPacketSocket * sock)
{
  if (sock->default_fd != -1 || sock->special_fd != NULL)
    return;

  close(sock->inet_fd);
  g_hash_table_remove(z_packsock_table, GINT_TO_POINTER(sock->port));
  z_io_watch_fd_destroy(sock->watch);
  g_free(sock);
}

/* Purges timed out ZPacketDispatcher's */

static void
z_packsock_filter_timeout(ZPacketSocket *sock)
{
  time_t now = time(NULL);
  ZPacketDispatcher **disp = &(sock->special_fd);

  while (*disp != NULL)
    {
      if ((*disp)->timeout < now)
	{
	  ZPacketDispatcher *tmp = *disp;
	  *disp = (*disp)->next;
	  z_sockaddr_unref(tmp->from);
	  z_sockaddr_unref(tmp->to);
	  g_free(tmp);
	  continue;		/* Because the step is done */
	}
      disp = &((*disp)->next);
    }
}

/* Received packet handler. Called by the z_io_watch* subsystem. */

static gboolean
z_packsock_packet(int fd, GIOCondition condition, gpointer user_data)
{
  ZPacketSocket *sock = (ZPacketSocket *) user_data;
  ZPacketDispatcher *disp;
  ZPacketDispatcher **disp_addr;

  struct sockaddr_in from;
  struct sockaddr_in *to_p = (struct sockaddr_in *) &(from.sin_zero);
  int from_len = sizeof(struct sockaddr_in);

  ZPacket packet;

  z_log(NULL, CORE_DEBUG, 8, "fd=%d, packet available", fd);

  packet.data = g_new(char, MAX_UDP_PACKET);

  /* IP: transparent proxy must be compiled into kernel */

  packet.length = recvfrom(sock->inet_fd, packet.data,
			   MAX_UDP_PACKET, MSG_PROXY, &from, &from_len);
  if (packet.length == -1)
    return TRUE;
  packet.data = g_renew(char, packet.data, packet.length);

  packet.from = z_sockaddr_new((struct sockaddr *) &from, sizeof(from));
  packet.to = z_sockaddr_new((struct sockaddr *) to_p,
			     sizeof(struct sockaddr_in));

  z_packsock_filter_timeout(sock);

  for (disp = sock->special_fd, disp_addr = &(sock->special_fd);
       disp != NULL; disp = disp->next, disp_addr = &(disp->next))
    {
      struct sockaddr_in *fromd, *tod;

      if (disp->from == NULL)
	fromd = NULL;
      else
	fromd = (struct sockaddr_in *) &disp->from->sa;

      if (disp->to == NULL)
	tod = NULL;
      else
	tod = (struct sockaddr_in *) &disp->to->sa;

      /* No check is made for the sockaddr being AF_INET */

      if (fromd != NULL &&
	  (fromd->sin_port != from.sin_port ||
	   fromd->sin_addr.s_addr != from.sin_addr.s_addr))
	continue;
      if (tod != NULL &&
	  (tod->sin_port != to_p->sin_port ||
	   fromd->sin_addr.s_addr != to_p->sin_addr.s_addr))
	continue;

      /* Packet must go to a special fd, send it and
         unlink that ZPacketDispatcher. */

      *disp_addr = disp->next;
      z_sockaddr_unref(disp->from);
      z_sockaddr_unref(disp->to);
      g_free(disp);

      write(disp->fd, &packet, sizeof(packet));
      return TRUE;

    }

  /* No special fd for this packet, send to default. */


  if (sock->default_fd != -1)
    {
      write(sock->default_fd, &packet, sizeof(packet));
      return TRUE;
    }
  {
    char buf[128], buf2[128];
    
    z_log(NULL, CORE_ERROR, 4, "fd=%d, internal error, got unexpected packet, from=%s to=%s",
              fd,
	      z_sockaddr_format(packet.from, buf, sizeof(buf)),
	      z_sockaddr_format(packet.to, buf2, sizeof(buf2)));
  }

  /* Packet is not sent, destroy all relevant data. */

  z_sockaddr_unref(packet.from);
  z_sockaddr_unref(packet.to);
  g_free(packet.data);

  return TRUE;
}

ZPacketSocket *
z_packsock_add_socket(int port)
{
  ZPacketSocket *sock = g_new0(ZPacketSocket, 1);
  struct sockaddr_in sin;
  int inet_fd;

  inet_fd = socket(AF_INET, SOCK_DGRAM, 0);
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if (bind(inet_fd, &sin, sizeof(sin)) < 0)
    {
      close(inet_fd);
      return NULL;
    }

  g_hash_table_insert(z_packsock_table, GINT_TO_POINTER(port), sock);
  sock->port = port;
  sock->default_fd = -1;
  sock->inet_fd = inet_fd;
  sock->special_fd = NULL;
  sock->watch = z_io_watch_fd_new(inet_fd,
				  G_IO_IN | G_IO_ERR | G_IO_PRI | G_IO_NVAL,
				  z_packsock_packet, sock);
  return sock;
}

int
z_packsock_add_default_fd(int port, int fd)
{
  ZPacketSocket *sock;

  z_log(NULL, CORE_DEBUG, 8, "z_packsock_add_default_fd(), port=%d, fd=%d", port, fd);
  z_packsock_init();
  sock = g_hash_table_lookup(z_packsock_table, GINT_TO_POINTER(port));
  if (sock == NULL)
    {
      sock = z_packsock_add_socket(port);
      if (sock == NULL)
	return 0;
    }
  z_packsock_filter_timeout(sock);
  if (sock->default_fd != -1)
    {
      return 0;
    }
  sock->default_fd = fd;
  return 1;
}

int
z_packsock_add_special_fd(int port,
			  ZSockAddr * from, ZSockAddr * to, int timeout,
			  int fd)
{
  ZPacketSocket *sock;
  ZPacketDispatcher *disp;

  if (from == NULL && to == NULL)
    return 0;
  if (timeout > time(NULL) + MAX_TIMEOUT || timeout < time(NULL))
    {
      z_log(NULL, CORE_ERROR, 1, "bad timeout requested for packet\n");
      return 0;
    }

  z_packsock_init();
  sock = g_hash_table_lookup(z_packsock_table, GINT_TO_POINTER(port));
  if (sock == NULL)
    {
      sock = z_packsock_add_socket(port);
      if (sock == NULL)
	return 0;
    }
  z_packsock_filter_timeout(sock);

  disp = g_new0(ZPacketDispatcher, 1);
  disp->from = from;
  disp->to = to;
  disp->fd = fd;
  disp->timeout = timeout;
  disp->next = sock->special_fd;
  sock->special_fd = disp;

  return 1;
}

void
z_packsock_close(int port)
{
  ZPacketSocket *sock;

  z_packsock_init();
  sock = g_hash_table_lookup(z_packsock_table, GINT_TO_POINTER(port));
  if (sock == NULL)
    return;
  sock->default_fd = -1;
  z_packsock_filter_timeout(sock);
  z_packsock_check_destroy(sock);
}

int
z_packsock_sendfrom(ZSockAddr * zfrom, ZSockAddr * zto, char *buf, int len)
{
  struct sockaddr_in to;
  struct sockaddr_in *from_p = (struct sockaddr_in *) &(to.sin_zero);
  struct sockaddr_in *zto_sa, *zfrom_sa;
  cap_t oldcaps;
  int rc;

  zto_sa = (struct sockaddr_in *) &zto->sa;
  zfrom_sa = (struct sockaddr_in *) &zfrom->sa;

  to.sin_family = AF_INET;
  to.sin_port = zto_sa->sin_port;
  to.sin_addr.s_addr = zto_sa->sin_addr.s_addr;

  from_p->sin_family = AF_INET;
  from_p->sin_port = zfrom_sa->sin_port;
  from_p->sin_addr.s_addr = zfrom_sa->sin_addr.s_addr;

  /* This is some deep kernel-level magic. Requires IP: transparent
     proxy to be compiled into kernel to work. */

  if (z_packsock_socket == -1)
    {
      z_packsock_socket = socket(AF_INET, SOCK_DGRAM, 0);
    }

  oldcaps = cap_save();
  cap_enable(CAP_NET_ADMIN);
  rc = sendto(z_packsock_socket, buf, len, MSG_PROXY, &to, sizeof(struct sockaddr_in));
  cap_restore(oldcaps);
  
  return rc;
}
