/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: stream.c,v 1.29.2.1 2001/04/06 14:56:02 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.9
 * Notes:
 *   - comments
 *
 ***************************************************************************/

#include <zorp/stream.h>
#include <zorp/log.h>

#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/poll.h>

void
z_data_dump(char *prefix, char *buf, guint len)
{
  guint i, j;
  gchar line[1024];
  
  i = 0;
  while (i < len)
    {
      char *end = line;
      
      for (j = 0; j < 16 && (i + j < len); j++)
        {
          g_snprintf(end, sizeof(line) - (end - line), "%02X ", (unsigned char) buf[i+j]);
          end += 3;
        }
      g_snprintf(end, sizeof(line) - (end - line), " ");
      end++;
      for (j = 0; j < 16 && (i + j < len) && sizeof(line) > (end - line); j++)
        {
          *end=buf[i+j]>31?buf[i+j]:'.';
          end++;
        }
      *end='\0';
          
      i += j;
      z_log(prefix, CORE_DEBUG, 9, "data line: %s", line);
    }
}

typedef ZStream ZRealStream;

static gboolean
z_stream_wait(ZStream *stream, guint cond, gint timeout)
{
  struct pollfd pfd;
  gint res;

  if (timeout == -2)
    return TRUE;
  errno = 0;
  pfd.fd = stream->fd;
  pfd.events = cond;
  pfd.revents = 0;
  do
    {
      res = poll(&pfd, 1, timeout);
      if (res == 1)
        {
          return TRUE;
        }
    }
  while (res == -1 && errno == EINTR);
  errno = ETIME;
  return FALSE;
}

static void 
z_stream_shutdown_fd(ZStream *s, int i)
{
  int res, attempt = 1, saved_errno;
  
  z_log(s->name, CORE_DEBUG, 7, "shutdown channel, mode=%d", i);
  
  errno = 0;
  do
    {
      res = shutdown(s->fd, i);
      saved_errno = errno;
      if (res == -1)
        {
          if (errno != EINTR)
            z_log(s->name, CORE_ERROR, 4, "shutdown failed, attempt=%d, %m", attempt++);
        }
    }
  while (saved_errno == EINTR);
}

static GIOError
z_stream_read_method(GIOChannel *channel,
		     gchar *buf,
		     guint count,
		     guint *bytes_read)
{
  ZRealStream *self = (ZRealStream *) channel;
  GIOError res;
  int attempt = 1, saved_errno;

  do
    { 
      if (!z_stream_wait(self, G_IO_IN | G_IO_HUP, self->timeout))
	{
	  z_log(self->name, CORE_ERROR, 1, "read timed out");
	  return G_IO_ERROR_UNKNOWN;
	}
      res = g_io_channel_read(self->channel, buf, count, bytes_read);
      saved_errno = errno;
      if (res != G_IO_ERROR_NONE)
        {
          if (errno != EINTR && errno != EAGAIN)
            z_log(self->name, CORE_ERROR, 1, "read failed, attempt=%d: %m", attempt++);
        }
    }
  while (res == G_IO_ERROR_AGAIN && saved_errno == EINTR);
  self->bytes_recvd += *bytes_read;

  z_log(self->name, CORE_DEBUG, 7, "reading channel, fd=%d, count=%d", self->fd, *bytes_read);
#if ENABLE_TRACE
  z_data_dump(self->name, buf, *bytes_read);
#endif
  return res;
}

static GIOError
z_stream_write_method(GIOChannel *channel,
		      gchar *buf,
		      guint count,
		      guint *bytes_written)
{
  ZRealStream *self = (ZRealStream *) channel;
  GIOError res;
  int attempt = 1, saved_errno;

  /* FIXME: todo, datadump */
  
  z_log(self->name, CORE_DEBUG, 7, "writing channel, fd=%d, count=%d", self->fd, count);

  do
    {
      if (!z_stream_wait(self, G_IO_OUT | G_IO_HUP, self->timeout))
	{
	  z_log(self->name, CORE_ERROR, 1, "write timed out");
	  return G_IO_ERROR_UNKNOWN;
	}
      res = g_io_channel_write(self->channel, buf, count, bytes_written);
      saved_errno = errno;
      if (res != G_IO_ERROR_NONE)
        {
	  if (errno != EINTR && errno != EAGAIN)
	    z_log(self->name, CORE_ERROR, 1, "write failed, attempt=%d: %m", attempt++);
        }
    }
  while (res == G_IO_ERROR_AGAIN && saved_errno == EINTR);
#if ENABLE_TRACE
  z_data_dump(self->name, buf, *bytes_written);
#endif
  self->bytes_sent += *bytes_written;
  return res;
}


GIOError
z_stream_send(ZStream *self, gchar *buf, guint buflen, int flags)
{
  int res, attempt = 1, saved_errno;
  
  errno = 0;
  do
    {
      if (!z_stream_wait(self, G_IO_OUT | G_IO_HUP, self->timeout))
	{
	  z_log(self->name, CORE_ERROR, 1, "send timed out");
	  return G_IO_ERROR_UNKNOWN;
	}
      res = send(self->fd, buf, buflen, flags);
      saved_errno = errno;
      if (res == -1)
        {
	  if (errno != EINTR && errno != EAGAIN)
	    z_log(self->name, CORE_ERROR, 1, "send failed, attempt=%d, %m", attempt++);
        }
    }
  while (res == -1 && saved_errno == EINTR);
  self->bytes_sent += buflen;
  if (res == 0)
    return G_IO_ERROR_NONE;
    
  if (saved_errno == EAGAIN)
    return G_IO_ERROR_AGAIN;
  else
    return G_IO_ERROR_UNKNOWN;
}


static void
z_stream_close_method(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;

  z_log(self->name, CORE_DEBUG, 7, "closing channel, fd=%d", self->fd);
  g_io_channel_close(self->channel);
}

static void
z_stream_free(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;
  time_t time_close;

  time_close = time(NULL);
  
  z_log(self->name, "core.accounting", 5, "accounting info, duration=%d, sent=%lld, received=%lld", 
        time_close - self->time_open, self->bytes_sent, self->bytes_recvd);
  g_io_channel_unref(self->channel);
  g_free(channel);
}

GIOFuncs z_stream_funcs =
{
  z_stream_read_method,
  z_stream_write_method,
  NULL,
  z_stream_close_method,
  NULL,
  z_stream_free
};

/*+

  Allocate and initialize a ZThread instance with the given fd and name.

  Parameters:
    fd        fd to wrap this stream around
    name      name to identify to stream in logs
  
  Returns:
    ZStream instance

  +*/
ZStream *
z_stream_new(gint fd, gchar *name)
{
  ZRealStream *self = g_new0(ZRealStream, 1);

  g_io_channel_init(&self->super);
  self->shutdown = z_stream_shutdown_fd;
  self->fd = fd;
  self->super.funcs = &z_stream_funcs;
  self->timeout = -2;
  self->time_open = time(NULL);
  self->channel = g_io_channel_unix_new(fd);
  strncpy(self->name, name, sizeof(self->name) - 1);
  return (ZStream *) self;
}

#if 0
void
z_stream_ssl(ZStream *s, SSL *ssl)
{
  ZRealStream *self = (ZRealStream *) s;
  
  g_io_channel_unref(self->channel);
  self->channel = g_io_channel_ssl_new(ssl);
}
#endif
