############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Ftp.py,v 1.10.2.3 2001/03/23 15:10:31 bazsi Exp $
##
## Author  : sasa
## Auditor :  
## Last version: 
## Notes:
##
############################################################################

"""Module exporting the Ftp proxy interface

This module defines the Ftp proxy interface as implemented by the
Ftp Zorp module.
"""

from Zorp import *
from Plug import PlugProxy
from Proxy import Proxy, proxyLog
from SockAddr import SockAddrInet, SockAddrInetRange
from Session import StackedSession
from Stream import Stream
import os         # for os.close

FTP_DATA_KEEP    = 0
FTP_DATA_PASSIVE = 1
FTP_DATA_ACTIVE  = 2

FTP_SIDE_CLIENT = 0
FTP_SIDE_SERVER = 1

FTP_CMD_DENY   = 0
FTP_CMD_ACCEPT = 1
FTP_CMD_POLICY = 2
FTP_CMD_ABORT  = 3

FTP_ANS_ACCEPT = 1
FTP_ANS_POLICY = 2
FTP_ANS_ABORT  = 3
FTP_ANS_CHANGE = 4

FTP_DEBUG  = "ftp.debug"
FTP_ERROR  = "ftp.error"
FTP_POLICY = "ftp.policy"

#class FtpDataConnect:
#	"""Helper class for FtpProxy to build data connections.
#
#	FtpDataConnect is a helper class for the Ftp proxy, it builds
#	the data-connection.
#	
#	Attributes
#	
#	"""



class FtpDataProxy(PlugProxy):
	"""Class encapsulating an FTP data connection.
	
	This class is used as a top-level proxy for FTP data connections.
	It is a simple plug (e.g. no content verification is done), but
	this class ensures that only one way traffic is transmitted
	by setting the appropriate copy_to_* attributes in Plug.
	
	"""
	def config(self):
		"""Configuration for FtpDataProxy
		
		This function sets the configuration for FtpDataProxy.  Our
		desired configuration is to restrict data transfer to one
		way only according to 'self.session.way'.
		
		Attributes
		
		  self -- this instance
		"""
		if self.session.way == 'C':
			self.copy_to_server = 0
		elif self.session.way == 'L':
			self.copy_to_client = 0
		else:
			proxyLog(self, FTP_ERROR, 4, "Unknown direction in FtpDataProxy, bidirectional dataconnection is used.")

	def __destroy__(self):
		"""Function called when the session is terminated.
		
		We call the registered stop() function to indicate the end
		of our session.
		
		Arguments
		
		  self -- this instance
		"""
		self.session.stop()

class FtpProxy(Proxy):
	"""Class encapsulating the FTP proxy implemented in Zorp.
	
	This proxy implements the FTP protocol. It denies everything by
	default, so uou should dervice your customized Ftp proxy classes in
	order to do something useful, or use 'FtpProxyAllow' or
	'FtpProxyMinimal'.

	Usage
	
	  All requests are denied in the low level proxy implementation by
	  default. To enable different commands and/or answers you'll need to
	  extend the low-level proxy in Python. To see how this is done,
	  see the next two sections.
	  
	  Setting policy for commands
	  
	    Changing the default behaviour of commands can be done using the
	    hash named 'self.command'. This hash is indexed by the command name
	    (e.g: USER or PWD), and each item in this hash is an action
	    tuple, defining proxy behaviour for the given command.
	    
	    The first item in this tuple is an integer value, determining
	    the action to take, and also the interpretation of the remaining
	    items in the tuple.
	  
	    Possible actions (Possible actions for commands in FTP):
	  
	      FTP_CMD_ACCEPT -- allow request without modification.

	      FTP_CMD_DENY   -- deny request. (default) Second parameter
	                        contains a string, which will be sent back to client.

	      FTP_CMD_POLICY -- call a Python function to decide what to do.
	                        this value uses an additional tuple item,
	                        which must be a callable Python function.
	                        The function takes two parameters:
	                        'self', 'command'

	      FTP_CMD_ABORT  -- deny request, and drop conection.
	  
	      Example (Sample for anonymous only sessions in FTP)

	        class AnonFtp(FtpProxy):

	          def config(self):
	            self.fw_server_data.ip_s="192.168.0.1"
                    self.fw_client_data.ip_s="10.0.0.1"
	            self.commands["USER"] = (Ftp.FTP_CMD_POLICY, self.pUser)
	            self.commands["*"] = (Ftp.FTP_CMD_ACCEPT)

	          def pUser(self,command):
	            if self.request_parameter == "ftp" or self.request_parameter == "anonymous":
	              return Z_ACCEPT
	            return Z_DENY

	  Changing answer code or string

	    Like with commands, we have a hash for answers as well.  This
	    hash indexed by command concatenated with answer, and contains an 
	    action tuple in each item. A special wildcard character '*' matches
	    every command. Another feature is that you don't need to write the
	    full answer code, it's possible to write only the first, or the
	    first two digits. 
	    
	    For example, all three index below is valid (Valid response codes in FTP):

	      PWD200  -- Match exactly the answer 200 coming in reply to a 
	                 PWD command.

	      PWD2    -- Match every answer starting with '2' in reply to a 
	                 PWD command.

	      *20     -- Match every answer between 200 and 209 in reply to 
	                 any command.

	    All answers are "denied" by default. It's changed to 
	    "500 Error parsing answer", because dropping an answer is not
	    permitted by RFC959. You may enable all answers with "*". 
	    
	    The precedence in processing hashtable entries is the following:

	      1. Exact match. (PWD200)
	        
	      2. Exact command match, partial answer matches (PWD20, PWD2, PWD)
	        
	      3. Wildcard command, with answer codes repeated as above. 
	         (*200, *20, *2, *)

	    Possible values for the first item in the action tuple (Possible action codes for responses in FTP):

	      FTP_ANS_ACCEPT -- allow answer without modification.

	      FTP_ANS_CHANGE -- change answer code and message. (default) 
	                        Second parameter contains a string, which 
	                        will be sent back to client.

	      FTP_ANS_POLICY -- call a Python function to decide what to do.
	                        this value uses an additional parameter,
	                        which must be a callable Python function.
	                        The function takes three parameters:
	                        self, command, answer

	      FTP_ANS_ABORT  -- deny request, and drop conection.
	
	Attributes 
	 
	  data_proxy              -- proxy to use for data connections
	  
	  data_port_min           -- ports should be allocated above this variable
	  
	  data_port_max           -- ports should be allocated below this variable
	  
	  data_mode               -- FTP_DATA_KEEP, FTP_DATA_PASSIVE or FTP_DATA_ACTIVE
	                             can be used to force a connection type in server
	                             side. (default: FTP_DATA_KEEP)

	  transparent_mode        -- (logical) TRUE for transparent proxy,
	                             FALSE otherwise (default: TRUE)
				     
	  fw_server_data          -- (ZorpSocket) Firewall address in server side.
	                             By default it contains the IP address of the
				     interface 'towards' the server.
				     
	  fw_client_data          -- (ZorpSocket) Firewall address in client side.
	                             By default it contains the IP address of the
				     interface 'towards' the client.
				     
	  permit_unknown_command  -- (boolean) Enable commands not understood
	                             (and not handled) by proxy. (Default: FALSE)
				     
	  permit_empty_command    -- (boolean) Enable lines without commands.
	                             (Default: FALSE)
				     
	  max_line_length         -- (integer) Maximum line lenght, what a proxy
	                             may transfer. (Default: 255)
				     
	  max_username_length     -- (integer) Maximum length of usernames.
	                             (Default: 32)
				     
	  max_password_length     -- (integer) Maximum length of passwords.
                                     (Default: 64)
				     
	  commands                -- (hash) normative policy hash, directing the
	                             proxy to do something with requests, without
	                             the need to call Python.  indexed by the
	                             command (e.g. "USER", "PWD" etc)
	                             (default: empty)
				     
	  answers                 -- (hash) normative policy hash directing the
	                             proxy to do something with answers. Indexed
	                             by the command, and the answer (e.g.
	                             "USER331", "PWD200" etc) (default: empty)
				     
	  request_command         -- (string) When a command goes up to policy level,
	                             you may change it with this.
	                             FIXME: Not yet used.
				     
	  request_parameter       -- (string) When a command goes up to policy level,
	                             this variable contains command parameters.
				     
	  answer_code             -- (string) When an answer goes up to policy level,
	                             this variable contains answer code.
				     
	  answer_parameter        -- (string) When an answer go up to policy level,
	                             this variable contain answer parameters.
				     
	  restrict_client_connect -- (boolean) Restricting data connection in
	                             client side. Accept data connection only if
	                             it's coming from IP number equal with command
	                             channel endpoint. (Default: TRUE)
				     
	  restrict_server_connect -- (boolean) Restricting data connection in
	                             server side. Accept data connection only if
	                             it's coming from IP number equal with command
	                             channel endpoint. (Default: FALSE)

	Private attributes
	
	  This attributes are used by the internal FTP logic, and shouldn't
	  be changed. It's documented here for completeness's sake.

	  state -- connection state 0 is initial, 1 one of the connections
	           were established, 2 both connections established ready to start data
	           proxy.
	           
	  modes -- array indexed by side, 'L' for Listen or 'C' for connect
	  
	  listens -- listen objects created by us, array indexed by side
	  
	  connects -- connect objects created by us, array indexed by side
	  
	  sides -- determines the order in which sides must be processed
	           (first listen/connect on sides[0] then on sides[1])
	  
	  fds -- array indexed by side, containing the side specific fd
	  
	  localsesa -- local socket adress in server side. Used when data
	               connections port is in a range.
	  
	  localclsa -- local socket adress in client side. Used when data
	               connecion port is in range.
	
	"""

	def __init__(self, session):
		"""Constructor to initialize an FtpProxy instance
		
		This constructor initializes an FtpProxy instance by
		calling the inherited __init__ constructor
		with appropriate parameters, and setting up
		local attributes based on arguments.
		
		Arguments

		  self          --      this instance
		  
		  session	--	The session object
		  
		"""
		Proxy.__init__(self, "ftp", session)
		self.restrict_client_connect = TRUE
		self.restrict_server_connect = FALSE
		self.data_proxy = FtpDataProxy
		self.data_port_min = 4000
		self.data_port_max = 6000
		self.state = 0
		self.modes = [' ', ' ']
		self.listens = [None, None]
		self.connects = [None, None]
		self.sides = [-1, -1]
		self.fds = [-1, -1]
		self.localsesa = None
		self.localclsa = None
	
	def __destroy__(self):
 		"""Function to deinitialize an FtpProxy class.
 		
 		This function is called by the Zorp core when the proxy is
 		to be freed. We call resetState() here to destroy bound
 		listeners and connectors.
 		
 		Arguments
 		
	 	  self -- this instance
 		"""
 		Proxy.__destroy__(self)
		self.resetData()

	def resetData(self):
 		"""Function reseting the data connection state.
 		
 		Sets the initial state of everything, and destroys pending
		listeners.
 		
 		Arguments
 		  
 		  self -- this instance
 		  
 		"""
		for i in range(0, 2):
			if self.listens[i] != None:
				self.listens[i].destroy()
			if self.connects[i] != None:
				self.connects[i].destroy()
		self.listens = [None, None]
		self.connects = [None, None]
		self.way = 'N'
		self.state = 0

	def stopDataConnection(self):
		"""Function called by the data proxy to indicate the end of the data connection.
		
		This callback is registered to the data proxy 
		('FtpDataProxy' by default), to be called when
		the data connection is terminated. It's task is
		to signal this condition to the Ftp proxy core.
		
		Arguments
		  
		  self -- this instance

		"""
		self.ftpDataWakeup('D')

	def prepareData(self, side, mode):
		"""Function to prepare the data channel on 'side' side using 'mode' mode.

		This function is called by the proxy core to prepare the
		given 'side' of the data connection according to 'mode'. 
		This means either to listen for connections (like passive
		mode on the client side) or to start establishing a
		connection (passive mode on the server side).

		Arguments
		
		  self  -- this instance
		  
		  side  -- FTP_SIDE_CLIENT or FTP_SIDE_SERVER representing
                           either the client or the server
                           
                  mode  -- 'L' or 'C' for Listen or Connect
                  
		"""
		if (side != FTP_SIDE_CLIENT and side != FTP_SIDE_SERVER):
			return 0

		if (side == FTP_SIDE_CLIENT):
			local = self.fw_client_data
			if local.ip == 0:
				local.ip = self.session.client_local.ip

			remote = self.client_data

			if local.port==0:
				if self.data_port_min == self.data_port_max == 0:
					self.localclsa = SockAddrInet(local.ip_s, 0)
				else:
					if self.localclsa == None:
						self.localclsa = SockAddrInetRange(local.ip_s, self.data_port_min, self.data_port_max)
				
				sa = self.localclsa
			else:
				sa = SockAddrInet(local.ip_s, local.port)
		
		elif (side == FTP_SIDE_SERVER):
			local = self.fw_server_data
			if local.ip == 0:
				local.ip = self.session.server_local.ip

			remote = self.server_data

			if self.data_port_min == self.data_port_max == 0:
				self.localsesa = SockAddrInet(local.ip_s, 0)
			else:
				if self.localsesa == None:
					self.localsesa = SockAddrInetRange(local.ip_s, self.data_port_min, self.data_port_max)
			sa = self.localsesa
			
		self.modes[side] = mode
		if (mode == 'L'):
			
			if self.listens[side] != None:
				self.listens[side].destroy()
				self.listens[side] = None
				proxyLog(self, FTP_ERROR, 4, "self.listen not None!")

			try:

				self.listens[side] = \
					     Listen(sa, 
						    self.acceptedCallback, 
					   	    1)
				if side == FTP_SIDE_CLIENT:
					self.fw_client_data = self.listens[side].local
				elif side == FTP_SIDE_SERVER:
					self.fw_server_data = self.listens[side].local
			except IOError:
				self.listens[side] = None
			
		elif (mode == 'C'):
			if self.connects[side] != None:
				self.connects[side].destroy()
				self.connects[side] = None
				proxyLog(self, FTP_ERROR, 4, "Self.connect not None!")
				
			try:
				self.connects[side] = \
					     Connect(sa, 
						     remote,
						     self.connectedCallback)
				if side == FTP_SIDE_CLIENT:
					self.fw_client_data = sa
				elif side == FTP_SIDE_SERVER:
					self.fw_server_data = sa
			except IOError:
				self.connects[side] = None

		return 0

	def startData(self, side1, side2, way):
		"""Function called by the proxy core to start the data connection procedure.
		
		This function starts the data connection proxy as specified 
		by previous calls to prepareData. We start the state machine
		here whose final state is to start a proxy on the established
		data connection.
		
		Arguments
		
		  side1  -- connect/listen on this side first
		  
		  side2  -- connect/listen on this side after the first one succeeded
		  
		  way    -- direction
		"""
		if side1 < 0 or side2 < 0 or side1 + side2 != 1:
			raise InternalError, "Side must be 0 or 1"
		else:
			if way != 'C' and way != 'S':
				raise InternalError, "Way must be C or S"
			else:
				self.sides = [side1, side2]
				self.state = 0
				self.way = way
				self.stepData()
                
	def stepData(self):
		"""Function called by the protocol logic to continue data establishment processing.
		
		This function is called each time after processing the
		current state finishes. This means that if the data
		connection on the server side is established we are
		ready to step one further and start accepting connections
		on the client side. 
		
		In state 0 and 1 it either listens or connects, based on
		the type field of parms[state].

		In state 2, stacks in the proxy and starts it with a
		session stacked out from self.session.

		Arguments
		
		  self -- this instance

		"""
		proxyLog(self, FTP_DEBUG, 7, "data listener, state %d" % (self.state))
		if self.state == 2:
        		# start proxy
			self.state = self.state + 1
			if self.fds[0] == 0 or self.fds[1] == 0:
				raise InternalError, "Any of fds is wrong!"
				self.resetstate()
				return

			proxyLog(self, FTP_DEBUG, 7, "stacking data proxy")
			
			# set up child session
			ss = StackedSession(self.session)
			ss.client_fd = self.fds[0]
			ss.client_stream = Stream(ss.client_fd, ss.session_id + 'data client')
			ss.server_fd = self.fds[1]
			ss.server_stream = Stream(ss.server_fd, ss.session_id + 'data server')
			ss.way = self.way
			ss.stop = self.stopDataConnection
			self.state = self.state + 1
			
			# start data proxy here
			self.data_proxy(ss)
			
			return

		elif self.state > 2 or self.state < 0:
			raise InternalError, "self.state greater than 2!"
			return

		side = self.sides[self.state]
		mode = self.modes[side]
		if mode == 'L':
			try:
				if self.listens[side] != None:
					self.listens[side].start()
			except IOError:
				self.listens[side] = None
		elif mode == 'C':
			try:
				if self.connects[side] != None:
					self.connects[side].start()
			except IOError:
				self.connects[side] = None
		else:
			proxyLog(self, FTP_ERROR, 2, "Internal error. Connection mode not L or C")

	def acceptedCallback(self, client, fd):
		"""Function called when a connection was accepted on either side.

		This callback is called if any of the sides was in 'L' mode,
		and a connection was accepted on the listening socket. After
		storing the parameters we continue processing by skipping to
		the next state.
		
		Arguments
		
		  self          -- this instance
		  
		  client	-- the client address which has connected
		  
		  fd		-- the file descriptor for the connection
		  
		"""

		proxyLog(self, FTP_DEBUG, 7, "data listener, acceptedCallback")
		
		if self.state > 1 or self.state < 0:
			raise InternalError, "Connection doesn't have more than two side!"
			return 0

		side = self.sides[self.state]
		self.state = self.state + 1
		self.fds[side] = fd
		if side == FTP_SIDE_CLIENT:

			if self.restrict_client_connect:
				# bounce check
				if client.ip != self.session.client_address.ip:
					os.close(fd)
					self.ftpDataWakeup('A')
					proxyLog(self, FTP_POLICY, 1, "Wrong IP when connecting in Client side.")
					return 0
			self.ftpDataWakeup('C')

		elif side == FTP_SIDE_SERVER:

			if self.restrict_server_connect:
				# bounce check
				if client.ip != self.session.server_address.ip:
					os.close(fd)
					self.ftpDataWakeup('B')
					proxyLog(self, FTP_POLICY, 1, "Wrong IP when connecting in server side")
					return 0
			self.ftpDataWakeup('S')
			
		else:
			proxyLog(self.ftp_state, FTP_ERROR, 4, "acceptedCallback called with bad state:%d" % (self.state-1))

 	def connectedCallback(self, fd):
		"""Function called when a connection was established on either side.
		
		This function is similar to 'acceptedCallback', but used if
		we were in 'C' (connect) mode. After storing the parameters
		we continue processing by skipping to the next state.
		
		Arguments
		
		  self  --      this instance
		  
		  fd	--	the file descriptor for the connection
		  
		"""
		proxyLog(self, FTP_DEBUG, 7, "data listener, connectedCallback")
		
		if self.state > 1 or self.state < 0:
			raise InternalError, "Connection doesn't have more than two side!"
			return

		side = self.sides[self.state]
		self.state = self.state + 1
		self.fds[side] = fd

		if side == FTP_SIDE_CLIENT:
			self.ftpDataWakeup('C')
		elif side == FTP_SIDE_SERVER:
			self.ftpDataWakeup('S')
		else:
			proxyLog(self, FTP_ERROR, 4,"connectedCallback called with bade state:%d" % (self.state-1))

	def loadAnswers(self):
		"""This function can be called by derived classes to initialize internal hashtables.

		This function fills in the self.answers hash so that commonly used
		request/answer combinations are enabled.

		Arguments

		  self -- this instance
		  
		"""
		self.answers["*421"] = (FTP_ANS_ACCEPT)
		self.answers["*500"] = (FTP_ANS_ACCEPT)
		self.answers["Null120"] = (FTP_ANS_ACCEPT)
		self.answers["Null220"] = (FTP_ANS_ACCEPT)
		self.answers["USER230"] = (FTP_ANS_ACCEPT)
		self.answers["USER530"] = (FTP_ANS_ACCEPT)
		self.answers["USER501"] = (FTP_ANS_ACCEPT)
		self.answers["USER331"] = (FTP_ANS_ACCEPT)
		self.answers["PASS230"] = (FTP_ANS_ACCEPT)
		self.answers["PASS202"] = (FTP_ANS_ACCEPT)
		self.answers["PASS530"] = (FTP_ANS_ACCEPT)
		self.answers["PASS501"] = (FTP_ANS_ACCEPT)
		self.answers["PASS503"] = (FTP_ANS_ACCEPT)
		self.answers["CWD250"]  = (FTP_ANS_ACCEPT)
		self.answers["CWD501"]  = (FTP_ANS_ACCEPT)
		self.answers["CWD502"]  = (FTP_ANS_ACCEPT)
		self.answers["CWD530"]  = (FTP_ANS_ACCEPT)
		self.answers["CWD550"]  = (FTP_ANS_ACCEPT)
		self.answers["CDUP200"] = (FTP_ANS_ACCEPT)
		self.answers["CDUP501"] = (FTP_ANS_ACCEPT)
		self.answers["CDUP502"] = (FTP_ANS_ACCEPT)
		self.answers["CDUP530"] = (FTP_ANS_ACCEPT)
		self.answers["CDUP550"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT202"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT250"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT501"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT502"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT530"] = (FTP_ANS_ACCEPT)
		self.answers["SMNT550"] = (FTP_ANS_ACCEPT)
		self.answers["REIN120"] = (FTP_ANS_ACCEPT)
		self.answers["REIN220"] = (FTP_ANS_ACCEPT)
		self.answers["REIN502"] = (FTP_ANS_ACCEPT)
		self.answers["QUIT221"] = (FTP_ANS_ACCEPT)
		self.answers["PORT200"] = (FTP_ANS_ACCEPT)
		self.answers["PORT501"] = (FTP_ANS_ACCEPT)
		self.answers["PORT530"] = (FTP_ANS_ACCEPT)
		self.answers["PASV227"] = (FTP_ANS_ACCEPT)
		self.answers["PASV501"] = (FTP_ANS_ACCEPT)
		self.answers["PASV502"] = (FTP_ANS_ACCEPT)
		self.answers["PASV530"] = (FTP_ANS_ACCEPT)
		self.answers["MODE200"] = (FTP_ANS_ACCEPT)
		self.answers["MODE501"] = (FTP_ANS_ACCEPT)
		self.answers["MODE504"] = (FTP_ANS_ACCEPT)
		self.answers["MODE530"] = (FTP_ANS_ACCEPT)
		self.answers["TYPE200"] = (FTP_ANS_ACCEPT)
		self.answers["TYPE501"] = (FTP_ANS_ACCEPT)
		self.answers["TYPE504"] = (FTP_ANS_ACCEPT)
		self.answers["TYPE530"] = (FTP_ANS_ACCEPT)
		self.answers["STRU200"] = (FTP_ANS_ACCEPT)
		self.answers["STRU501"] = (FTP_ANS_ACCEPT)
		self.answers["STRU504"] = (FTP_ANS_ACCEPT)
		self.answers["STRU530"] = (FTP_ANS_ACCEPT)
		self.answers["ALLO200"] = (FTP_ANS_ACCEPT)
		self.answers["ALLO202"] = (FTP_ANS_ACCEPT)
		self.answers["ALLO501"] = (FTP_ANS_ACCEPT)
		self.answers["ALLO504"] = (FTP_ANS_ACCEPT)
		self.answers["ALLO530"] = (FTP_ANS_ACCEPT)
		self.answers["REST501"] = (FTP_ANS_ACCEPT)
		self.answers["REST502"] = (FTP_ANS_ACCEPT)
		self.answers["REST530"] = (FTP_ANS_ACCEPT)
		self.answers["REST350"] = (FTP_ANS_ACCEPT)
		self.answers["STOR125"] = (FTP_ANS_ACCEPT)
		self.answers["STOR150"] = (FTP_ANS_ACCEPT)
		self.answers["STOR110"] = (FTP_ANS_ACCEPT)
		self.answers["STOR226"] = (FTP_ANS_ACCEPT)
		self.answers["STOR250"] = (FTP_ANS_ACCEPT)
		self.answers["STOR425"] = (FTP_ANS_ACCEPT)
		self.answers["STOR426"] = (FTP_ANS_ACCEPT)
		self.answers["STOR451"] = (FTP_ANS_ACCEPT)
		self.answers["STOR551"] = (FTP_ANS_ACCEPT)
		self.answers["STOR552"] = (FTP_ANS_ACCEPT)
		self.answers["STOR532"] = (FTP_ANS_ACCEPT)
		self.answers["STOR450"] = (FTP_ANS_ACCEPT)
		self.answers["STOR452"] = (FTP_ANS_ACCEPT)
		self.answers["STOR553"] = (FTP_ANS_ACCEPT)
		self.answers["STOR501"] = (FTP_ANS_ACCEPT)
		self.answers["STOR530"] = (FTP_ANS_ACCEPT)
		self.answers["STOU125"] = (FTP_ANS_ACCEPT)
		self.answers["STOU150"] = (FTP_ANS_ACCEPT)
		self.answers["STOU110"] = (FTP_ANS_ACCEPT)
		self.answers["STOU226"] = (FTP_ANS_ACCEPT)
		self.answers["STOU250"] = (FTP_ANS_ACCEPT)
		self.answers["STOU425"] = (FTP_ANS_ACCEPT)
		self.answers["STOU426"] = (FTP_ANS_ACCEPT)
		self.answers["STOU451"] = (FTP_ANS_ACCEPT)
		self.answers["STOU551"] = (FTP_ANS_ACCEPT)
		self.answers["STOU552"] = (FTP_ANS_ACCEPT)
		self.answers["STOU532"] = (FTP_ANS_ACCEPT)
		self.answers["STOU450"] = (FTP_ANS_ACCEPT)
		self.answers["STOU452"] = (FTP_ANS_ACCEPT)
		self.answers["STOU553"] = (FTP_ANS_ACCEPT)
		self.answers["STOU501"] = (FTP_ANS_ACCEPT)
		self.answers["STOU530"] = (FTP_ANS_ACCEPT)
		self.answers["RETR125"] = (FTP_ANS_ACCEPT)
		self.answers["RETR150"] = (FTP_ANS_ACCEPT)
		self.answers["RETR110"] = (FTP_ANS_ACCEPT)
		self.answers["RETR226"] = (FTP_ANS_ACCEPT)
		self.answers["RETR250"] = (FTP_ANS_ACCEPT)
		self.answers["RETR425"] = (FTP_ANS_ACCEPT)
		self.answers["RETR426"] = (FTP_ANS_ACCEPT)
		self.answers["RETR451"] = (FTP_ANS_ACCEPT)
		self.answers["RETR532"] = (FTP_ANS_ACCEPT)
		self.answers["RETR450"] = (FTP_ANS_ACCEPT)
		self.answers["RETR452"] = (FTP_ANS_ACCEPT)
		self.answers["RETR553"] = (FTP_ANS_ACCEPT)
		self.answers["RETR501"] = (FTP_ANS_ACCEPT)
		self.answers["RETR530"] = (FTP_ANS_ACCEPT)
		self.answers["LIST125"] = (FTP_ANS_ACCEPT)
		self.answers["LIST150"] = (FTP_ANS_ACCEPT)
		self.answers["LIST226"] = (FTP_ANS_ACCEPT)
		self.answers["LIST250"] = (FTP_ANS_ACCEPT)
		self.answers["LIST425"] = (FTP_ANS_ACCEPT)
		self.answers["LIST426"] = (FTP_ANS_ACCEPT)
		self.answers["LIST451"] = (FTP_ANS_ACCEPT)
		self.answers["LIST450"] = (FTP_ANS_ACCEPT)
		self.answers["LIST501"] = (FTP_ANS_ACCEPT)
		self.answers["LIST502"] = (FTP_ANS_ACCEPT)
		self.answers["LIST530"] = (FTP_ANS_ACCEPT)
		self.answers["NLST125"] = (FTP_ANS_ACCEPT)
		self.answers["NLST150"] = (FTP_ANS_ACCEPT)
		self.answers["NLST226"] = (FTP_ANS_ACCEPT)
		self.answers["NLST250"] = (FTP_ANS_ACCEPT)
		self.answers["NLST425"] = (FTP_ANS_ACCEPT)
		self.answers["NLST426"] = (FTP_ANS_ACCEPT)
		self.answers["NLST451"] = (FTP_ANS_ACCEPT)
		self.answers["NLST450"] = (FTP_ANS_ACCEPT)
		self.answers["NLST501"] = (FTP_ANS_ACCEPT)
		self.answers["NLST502"] = (FTP_ANS_ACCEPT)
		self.answers["NLST530"] = (FTP_ANS_ACCEPT)
		self.answers["APPE125"] = (FTP_ANS_ACCEPT)
		self.answers["APPE150"] = (FTP_ANS_ACCEPT)
		self.answers["APPE110"] = (FTP_ANS_ACCEPT)
		self.answers["APPE226"] = (FTP_ANS_ACCEPT)
		self.answers["APPE250"] = (FTP_ANS_ACCEPT)
		self.answers["APPE425"] = (FTP_ANS_ACCEPT)
		self.answers["APPE426"] = (FTP_ANS_ACCEPT)
		self.answers["APPE451"] = (FTP_ANS_ACCEPT)
		self.answers["APPE551"] = (FTP_ANS_ACCEPT)
		self.answers["APPE552"] = (FTP_ANS_ACCEPT)
		self.answers["APPE532"] = (FTP_ANS_ACCEPT)
		self.answers["APPE450"] = (FTP_ANS_ACCEPT)
		self.answers["APPE550"] = (FTP_ANS_ACCEPT)
		self.answers["APPE452"] = (FTP_ANS_ACCEPT)
		self.answers["APPE553"] = (FTP_ANS_ACCEPT)
		self.answers["APPE501"] = (FTP_ANS_ACCEPT)
		self.answers["APPE502"] = (FTP_ANS_ACCEPT)
		self.answers["APPE530"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR450"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR550"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR501"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR502"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR530"] = (FTP_ANS_ACCEPT)
		self.answers["RNFR350"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO250"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO532"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO553"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO501"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO502"] = (FTP_ANS_ACCEPT)
		self.answers["RNTO530"] = (FTP_ANS_ACCEPT)
		self.answers["DELE250"] = (FTP_ANS_ACCEPT)
		self.answers["DELE450"] = (FTP_ANS_ACCEPT)
		self.answers["DELE550"] = (FTP_ANS_ACCEPT)
		self.answers["DELE501"] = (FTP_ANS_ACCEPT)
		self.answers["DELE502"] = (FTP_ANS_ACCEPT)
		self.answers["DELE530"] = (FTP_ANS_ACCEPT)
		self.answers["RMD250"]  = (FTP_ANS_ACCEPT)
		self.answers["RMD501"]  = (FTP_ANS_ACCEPT)
		self.answers["RMD502"]  = (FTP_ANS_ACCEPT)
		self.answers["RMD530"]  = (FTP_ANS_ACCEPT)
		self.answers["RMD550"]  = (FTP_ANS_ACCEPT)
		self.answers["MKD257"]  = (FTP_ANS_ACCEPT)
		self.answers["MKD501"]  = (FTP_ANS_ACCEPT)
		self.answers["MKD502"]  = (FTP_ANS_ACCEPT)
		self.answers["MKD530"]  = (FTP_ANS_ACCEPT)
		self.answers["MKD550"]  = (FTP_ANS_ACCEPT)
		self.answers["PWD257"]  = (FTP_ANS_ACCEPT)
		self.answers["PWD501"]  = (FTP_ANS_ACCEPT)
		self.answers["PWD502"]  = (FTP_ANS_ACCEPT)
		self.answers["PWD550"]  = (FTP_ANS_ACCEPT)
		self.answers["ABOR225"] = (FTP_ANS_ACCEPT)
		self.answers["ABOR226"] = (FTP_ANS_ACCEPT)
		self.answers["ABOR501"] = (FTP_ANS_ACCEPT)
		self.answers["ABOR502"] = (FTP_ANS_ACCEPT)
		self.answers["SYST215"] = (FTP_ANS_ACCEPT)
		self.answers["SYST501"] = (FTP_ANS_ACCEPT)
		self.answers["SYST502"] = (FTP_ANS_ACCEPT)
		self.answers["STAT211"] = (FTP_ANS_ACCEPT)
		self.answers["STAT212"] = (FTP_ANS_ACCEPT)
		self.answers["STAT213"] = (FTP_ANS_ACCEPT)
		self.answers["STAT450"] = (FTP_ANS_ACCEPT)
		self.answers["STAT501"] = (FTP_ANS_ACCEPT)
		self.answers["STAT502"] = (FTP_ANS_ACCEPT)
		self.answers["STAT530"] = (FTP_ANS_ACCEPT)
		self.answers["HELP211"] = (FTP_ANS_ACCEPT)
		self.answers["HELP214"] = (FTP_ANS_ACCEPT)
		self.answers["HELP501"] = (FTP_ANS_ACCEPT)
		self.answers["HELP502"] = (FTP_ANS_ACCEPT)
		self.answers["SITE200"] = (FTP_ANS_ACCEPT)
		self.answers["SITE202"] = (FTP_ANS_ACCEPT)
		self.answers["SITE501"] = (FTP_ANS_ACCEPT)
		self.answers["SITE530"] = (FTP_ANS_ACCEPT)
		self.answers["NOOP200"] = (FTP_ANS_ACCEPT)
		self.answers["SIZE213"] = (FTP_ANS_ACCEPT)
		self.answers["MDTM213"] = (FTP_ANS_ACCEPT)
		self.answers["SIZE550"] = (FTP_ANS_ACCEPT)
		self.answers["MDTM550"] = (FTP_ANS_ACCEPT)

class FtpProxyAllow(FtpProxy):
	"""Class encapsulating a permitting ftp proxy which allows everything, even unknown commands.
	    
	This class encapsulates a permitting Ftp proxy allowing unknown commands
	and unknown answers.
	"""

	def config(self):
		"""Configuration for FtpProxyAllow.
		
		Enables all commands by setting permit_unknown_commands to
		TRUE and adding two wildcard entries to the commands
		and answers hash.

		Arguments
		
		  self -- this instance
		
		"""
		self.commands["*"] = (FTP_CMD_ACCEPT)
		self.answers["*"]  = (FTP_CMD_ACCEPT)
		self.permit_unknown_command = TRUE

class FtpProxyMinimal(FtpProxy):
	"""Class encapsulating a minimalist FTP proxy which allows minimal set of commands.

	This proxy enables the minimal required protocol elements, and strictly
	checks commands and return codes. For unknown commands or replies
	the session is terminated.
	"""
	
	def test(self, command, answer):
		"""Test function to log answers not enabled in config

		This function is used to gather still useful answers
		not enabled in FtpProxyMinimal. It denies all
		unknown answers but sends a log message about it.

		Arguments
		
		  self -- this instance
		  
		  command -- command

		  answer -- answer

		"""
		proxyLog(self, FTP_DEBUG, 6, "Command send: %s, Answer code: %s" % (command, answer))
		return Z_ABORT

	def config(self):
		"""Configuration for FtpProxyMinimal

		It enables a minimal set of commands for a working FTP proxy, and
		sets permit_unknown_commands to FALSE.
		
		Arguments
		
		  self -- this instance
		
		"""
		self.commands["ABOR"] = (FTP_CMD_ACCEPT)
		self.commands["CDUP"] = (FTP_CMD_ACCEPT)
		self.commands["CLNT"] = (FTP_CMD_DENY)
		self.commands["CWD"]  = (FTP_CMD_ACCEPT)
		self.commands["FEAT"] = (FTP_CMD_DENY)
		self.commands["LIST"] = (FTP_CMD_ACCEPT)
		self.commands["MDTM"] = (FTP_CMD_ACCEPT)
		self.commands["MODE"] = (FTP_CMD_ACCEPT)
		self.commands["NLST"] = (FTP_CMD_ACCEPT)
		self.commands["NOOP"] = (FTP_CMD_ACCEPT)
		self.commands["PASV"] = (FTP_CMD_ACCEPT)
		self.commands["PASS"] = (FTP_CMD_ACCEPT)
		self.commands["PORT"] = (FTP_CMD_ACCEPT)
		self.commands["PWD"]  = (FTP_CMD_ACCEPT)
		self.commands["QUIT"] = (FTP_CMD_ACCEPT)
		self.commands["REST"] = (FTP_CMD_ACCEPT)
		self.commands["RETR"] = (FTP_CMD_ACCEPT)
		self.commands["RNFR"] = (FTP_CMD_ACCEPT)
		self.commands["RNTO"] = (FTP_CMD_ACCEPT)
		self.commands["SIZE"] = (FTP_CMD_ACCEPT)
		self.commands["STOR"] = (FTP_CMD_ACCEPT)
		self.commands["STAT"] = (FTP_CMD_ACCEPT)
		self.commands["STRU"] = (FTP_CMD_ACCEPT)
		self.commands["TYPE"] = (FTP_CMD_ACCEPT)
		self.commands["USER"] = (FTP_CMD_ACCEPT)
		self.commands["XPWD"] = (FTP_CMD_DENY)
		self.commands["MACB"] = (FTP_CMD_DENY)  #Who was that &^%$$^&, that not in any rfcs.

		FtpProxy.loadAnswers(self)
		self.answers["*"] = (FTP_CMD_POLICY, self.test)
		self.permit_unknown_command = FALSE
