/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: http.c,v 1.40.2.5 2001/04/06 14:56:02 bazsi Exp $
 * 
 * Author: Balazs Scheidler <bazsi@balabit.hu>
 * Auditor: 
 * Last audited version: 
 * Notes:
 *   Based on the code by: Viktor Peter Kovacs <vps__@freemail.hu>
 *   
 ***************************************************************************/

#include "http.h"

#include <zorp/thread.h>
#include <zorp/registry.h>
#include <zorp/log.h>
#include <zorp/policy.h>

#include <ctype.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

#define HTTP_DEBUG "http.debug"
#define HTTP_ERROR "http.error"
#define HTTP_POLICY "http.policy"
#define HTTP_REQUEST "http.request"
#define HTTP_RESPONSE "http.response"

/* case insensitive compare */
static int
http_filter_hash_compare(gconstpointer a, gconstpointer b)
{
  if (strcasecmp((char *) a, (char *) b) == 0)
    {
      return (1);
    }
  return (0);
}

static int
http_filter_hash_bucket(gconstpointer a)
{
  int sum = 0;
  char *s = (char *) a;
  
  while (*s != 0)
    {
      sum += toupper(*s);     
      s++;
    }
  return (sum % 16);          
}

static void
http_config_set_defaults(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  self->connection_mode = HTTP_CONNECTION_CLOSE;
  self->transparent_mode = TRUE;
  self->transparent_server_requests = TRUE;
  self->transparent_proxy_requests = FALSE;

  self->max_line_length = 4096;
  self->max_header_lines = 50;
  self->timeout = 30000;
  self->default_port = 80;

  self->headers[EP_CLIENT] = g_string_sized_new(256);
  self->headers[EP_SERVER] = g_string_sized_new(256);

  self->request_url = g_string_sized_new(HTTP_MAX_URL);
  self->current_header_name = g_string_sized_new(16);
  self->current_header_value = g_string_sized_new(32);

  self->parent_proxy = g_string_sized_new(0);
  self->parent_proxy_port = 3128;

  self->request_method_policy =
    g_hash_table_new((GHashFunc) http_filter_hash_bucket,
                     (GCompareFunc) http_filter_hash_compare);
  self->request_header_policy =
    g_hash_table_new((GHashFunc) http_filter_hash_bucket,
                     (GCompareFunc) http_filter_hash_compare);
  self->response_policy =
    g_hash_table_new((GHashFunc) http_filter_hash_bucket,
                     (GCompareFunc) http_filter_hash_compare);
  self->response_header_policy =
    g_hash_table_new((GHashFunc) http_filter_hash_bucket,
                     (GCompareFunc) http_filter_hash_compare);


  self->error_info = g_string_sized_new(0);
  self->error_code = -1;
  self->error_response = 500;
  z_func_leave(self, HTTP_DEBUG);
}

static void
http_config_init(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  if (self->max_line_length > HTTP_MAX_LINE)
    self->max_line_length = HTTP_MAX_LINE;
  self->super.endpoints[EP_CLIENT]->timeout = self->timeout;
  z_func_leave(self, HTTP_DEBUG);
}

static void
http_register_vars(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  /* yes, no */
  z_proxy_var_new(&self->super, "transparent_mode",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG,
                  &self->transparent_mode);

  /* yes, no */
  z_proxy_var_new(&self->super, "transparent_server_requests",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
                  &self->transparent_server_requests);

  /* yes, no */
  z_proxy_var_new(&self->super, "transparent_proxy_requests",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
                  &self->transparent_proxy_requests);

  /* close or keep-alive */
  z_proxy_var_new(&self->super, "connection_mode", 
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET,
                  &self->connection_mode);

  /* string containing parent proxy */
  z_proxy_var_new(&self->super, "parent_proxy",
		  Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
		  self->parent_proxy);

  /* parent proxy port */
  z_proxy_var_new(&self->super, "parent_proxy_port",
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
		  &self->parent_proxy_port);

  /* default port if portnumber is not specified in urls */
  z_proxy_var_new(&self->super, "default_port",
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
		  &self->default_port);

  /* rewrite host header when redirecting */
  z_proxy_var_new(&self->super, "rewrite_host_header",
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
		  &self->rewrite_host_header);

  /* integer */
  z_proxy_var_new(&self->super, "max_line_length",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
                  &self->max_line_length);

  /* integer */
  z_proxy_var_new(&self->super, "max_header_lines",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
                  &self->max_header_lines);

  /* integer */
  z_proxy_var_new(&self->super, "max_keepalive_requests",
                  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG,
                  &self->max_keepalive_requests);

  /* timeout value in milliseconds */
  z_proxy_var_new(&self->super, "timeout", 
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_GET_CONFIG | Z_VAR_SET_CONFIG,
                  &self->timeout);

  /* hash indexed by request method */
  z_proxy_var_new(&self->super, "request",
                  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG,
                  self->request_method_policy);

  /* hash indexed by header name */
  z_proxy_var_new(&self->super, "request_headers",
                  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG,
                  self->request_header_policy);

  /* hash indexed by response code */
  /* FIXME: not yet used */
  z_proxy_var_new(&self->super, "response",
                  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG,
                  self->response_policy);

  /* hash indexed by header name */
  z_proxy_var_new(&self->super, "response_headers",
                  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG,
                  self->response_header_policy);
  
  /* string containing current url */
  z_proxy_var_new(&self->super, "request_url",
		  Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET,
		  self->request_url);

  /* string containing current header name */
  z_proxy_var_new(&self->super, "current_header_name",
		  Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET,
		  self->current_header_name);

  /* string containing current header value */
  z_proxy_var_new(&self->super, "current_header_value",
		  Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET,
		  self->current_header_value);

  /* error response */
  z_proxy_var_new(&self->super, "error_response", 
		  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET,
                  &self->error_response);

  /* string inserted into error messages */
  z_proxy_var_new(&self->super, "error_info",
		  Z_VAR_TYPE_STRING | Z_VAR_GET | Z_VAR_SET,
		  self->error_info);

  z_func_leave(self, HTTP_DEBUG);
}

static gboolean
http_error_message(HttpProxy *self, int response_code, int message_code, GString *infomsg)
{
  gchar *messages[] = { NULL, "clientsyntax.html", "serversyntax.html", "policysyntax.html", "policyviolation.html", "invalidurl.html", "connecterror.html", "ioerror.html"};
  gchar response[64], filename[256];
  int fd;
 
  z_func_enter(self, HTTP_DEBUG); 
  if (message_code < 0 || (message_code >= sizeof(messages) / sizeof(char *)))
    {
      /* message code out of range */
      z_proxy_log(self, HTTP_ERROR, 1, "Internal error, message code out of range. message_code=%d", message_code);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
    
  if (message_code == 0)
    {
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    }
    
  if (self->proto_version >= 0x0100)
    {
      g_snprintf(response, sizeof(response), "HTTP/1.0 %d Proxy error encountered\r\n", response_code);
      
      if (!http_write(self, EP_CLIENT, response, strlen(response)))
        {
	/* error writing error message */
	  z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
      g_string_truncate(self->headers[EP_SERVER], 0);
      if (!self->transparent_mode)
        g_string_append(self->headers[EP_SERVER], "Proxy-Connection: close\r\n");
      else
        g_string_append(self->headers[EP_SERVER], "Connection: close\r\n");
      
      g_string_append(self->headers[EP_SERVER], "Content-Type: text/html\r\n");
      g_string_append(self->headers[EP_SERVER], "\r\n");
      if (!http_write(self, EP_CLIENT, self->headers[EP_SERVER]->str, self->headers[EP_SERVER]->len))
        {
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
    }
  g_snprintf(filename, sizeof(filename), "%s/http/%s", ZORP_DATADIR, messages[message_code]);
  z_proxy_log(self, HTTP_DEBUG, 5, "error, serving %s", filename);
  fd = open(filename, O_RDONLY);
  if (fd == -1)
    {
      z_proxy_log(self, HTTP_ERROR, 3, "error opening error file %s", filename);
      http_write(self, EP_CLIENT, "Proxy error encountered, error file not found.\r\n", 49);
    }
  else
    {
      gchar contents[4096], *info;
      gint count;
      
      count = read(fd, contents, sizeof(contents) - 1);
      close(fd);
      if (count == -1)
	count = 0;
      contents[count] = 0;
      info = strstr(contents, "@INFO@");
      if (info)
	{
	  GString *new_contents = g_string_sized_new(4096);
	  
	  g_string_append_len(new_contents, contents, info - contents - 1);
	  if (infomsg)
	    g_string_append_len(new_contents, infomsg->str, infomsg->len);
	  g_string_append(new_contents, "<hr>\nPage generated by Zorp version ");
	  g_string_append(new_contents, VERSION);
	  g_string_append(new_contents, "<br>\n");
	  g_string_append(new_contents, info + 6);
	  http_write(self, EP_CLIENT, new_contents->str, new_contents->len);
	  g_string_free(new_contents, 1);
	}
      else
	http_write(self, EP_CLIENT, contents, count);
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_client_stream_init(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  self->readline[EP_CLIENT] = z_read_line_new(self->super.endpoints[EP_CLIENT], self->max_line_length, ZRL_EOL_CRLF | ZRL_TRUNCATE);
  /* timeout is initialized after the config event */
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_server_stream_init(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  self->readline[EP_SERVER] = z_read_line_new(self->super.endpoints[EP_SERVER], self->max_line_length, ZRL_EOL_CRLF | ZRL_TRUNCATE);
  self->super.endpoints[EP_SERVER]->timeout = self->timeout;
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

gboolean
http_read(HttpProxy *self, int side, gchar *buf, size_t buflen, guint *bytes_read)
{
  guint res;
  
  z_func_enter(self, HTTP_DEBUG);
  /* try to fetch data already buffered in readline */
  if (z_read_line_has_unread_data(self->readline[side]))
    {
      res = z_read_line_get_unread_data(self->readline[side], buf, buflen, bytes_read);
      if (ST_ERROR(res) != ST_OK)
        {
          z_proxy_log(self, HTTP_ERROR, 1, "Error fetching already buffered data, res=%x", res);
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    }
  /* no data in readline, fetch from the stream */
  res = z_stream_read(self->super.endpoints[side], buf, buflen, bytes_read);
  if (ST_ERROR(res) != ST_OK)
    {
      z_proxy_log(self, HTTP_ERROR, 1, "Error reading %s stream, res=%x, error=%m", side == EP_CLIENT ? "client" : "server", res);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

gboolean
http_write(HttpProxy *self, int side, gchar *buf, size_t buflen)
{
  guint res, bytes_written;
  
  z_func_enter(self, HTTP_DEBUG);
  if (!self->super.endpoints[side])
    {
      /* write to NULL stream */
      z_proxy_log(self, HTTP_ERROR, 1, "Error writing %s stream, stream is NULL", side == EP_CLIENT ? "client" : "server");
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  res = z_stream_write(self->super.endpoints[side], buf, buflen, &bytes_written);
  if (ST_ERROR(res) != ST_OK || buflen != bytes_written)
    {
      /* FIXME: move this to a separate function */
      z_proxy_log(self, HTTP_ERROR, 1, "Error writing %s stream, res=%x, error=%m", side == EP_CLIENT ? "client" : "server", res);

      z_stream_shutdown(self->super.endpoints[side], SHUT_RDWR);
      z_stream_close(self->super.endpoints[side]);
      z_stream_unref(self->super.endpoints[side]);
      z_read_line_destroy(self->readline[side]);
      self->super.endpoints[side] = NULL;
      self->readline[side] = NULL;
      
      self->error_code = HTTP_MSG_IO_ERROR;
      self->error_response = 502;
      g_string_sprintf(self->error_info, "Error writing to %s (%m)", side == EP_CLIENT ? "client" : "server");
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}   

static gboolean
http_hash_get_type(ZPolicyObj *tuple, guint *filter_type)
{
  ZPolicyObj *tmp;

  if (!z_policy_seq_check(tuple))
    {
      if (z_policy_var_parse(tuple, "i", filter_type))
        {
          return TRUE;
        }
      /* not a sequence nor an int */
      return FALSE;
    }
    
  tmp = z_policy_seq_getitem(tuple, 0);
  if (!z_policy_var_parse(tmp, "i", filter_type))
    {
      /* policy syntax error */
      z_policy_var_unref(tmp);
      return FALSE;
    }
  z_policy_var_unref(tmp);
  return TRUE;
}


#define SKIP_SPACES \
  do \
    { \
      while (left > 0 && *src == ' ') \
        { \
          src++; \
          left--; \
        } \
    } \
  while (0)
  
#define COPY_SPACE \
  do \
    { \
      while (left > 0 && avail > 0 && *src != ' ' && *src) \
        { \
          *dst++ = *src++; \
          left--; \
          avail--; \
        } \
      *dst = 0; \
    } \
  while (0)
	

static gboolean
http_split_request(HttpProxy *self, gchar *line, guint length)
{
  gchar *src, *dst;
  gint left, avail;

  z_func_enter(self, HTTP_DEBUG);
  self->request_method[0] = 0;
  self->request_flags = -1;
  self->request_version[0] = 0;

  g_string_truncate(self->request_url, 0);

  src = line;
  left = length;
  dst = self->request_method;
  avail = sizeof(self->request_method) - 1;

  COPY_SPACE;

  if (!self->request_method[0] || (*src != ' ' && avail == 0))
    {
      z_func_leave(self, HTTP_DEBUG);
      /* request method empty, or request buffer overflow */
      return FALSE;
    }

  SKIP_SPACES;

  avail = HTTP_MAX_URL;
  g_string_truncate(self->request_url, 0);

  while (left > 0 && avail > 0 && *src != ' ' && *src) 
    { 
      g_string_append_c(self->request_url, *src++);
      left--; 
      avail--; 
    } 

  if (!self->request_url->str[0] || (*src != ' ' && avail == 0))
    {
      /* url missing, or too long */
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  SKIP_SPACES;

  dst = self->request_version;
  avail = sizeof(self->request_version) - 1;
  
  COPY_SPACE;

  if (*src != ' ' && avail == 0)
    {
      /* protocol version too long */
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  z_proxy_log(self, HTTP_REQUEST, 7, "Request: command=%s, url=%s, version=%s", self->request_method, self->request_url->str, self->request_version);
  z_func_leave(self, HTTP_DEBUG);
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_split_response(HttpProxy *self, gchar *line, guint line_length, gboolean *reply09)
{
  gchar *src, *dst;
  gint left, avail;

  z_func_enter(self, HTTP_DEBUG);
  self->response_version[0] = 0;
  self->response[0] = 0;
  self->response_msg[0] = 0;
  *reply09 = FALSE;
  
  src = line;
  left = line_length;
  
  dst = self->response_version;
  avail = sizeof(self->response_version) - 1;

  COPY_SPACE;
  
  if (memcmp(self->response_version, "HTTP", 4) != 0)
    {
      /* no status line */
      *reply09 = TRUE;
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if (!self->response_version[0] || (*src != ' ' && avail == 0))
    {
      /* response version empty or too long */
      z_proxy_log(self, HTTP_REQUEST, 3, "response version empty or too long, line=%.*s", line_length, line);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  
  SKIP_SPACES;
  
  dst = self->response;
  avail = sizeof(self->response) - 1;

  COPY_SPACE;
  
  if (!self->response[0] || (*src != ' ' && left && avail == 0))
    {
      /* response code empty or too long */
      z_proxy_log(self, HTTP_REQUEST, 3, "response code empty or too long, line=%.*s", line_length, line);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  
  self->response_code = atoi(self->response);
    
  SKIP_SPACES;

  dst = self->response_msg;
  avail = sizeof(self->response_msg) - 1;

  while (left > 0 && avail > 0) 
    { 
      *dst++ = *src++; 
      left--; 
      avail--; 
    } 
  *dst = 0; 
  
  z_proxy_log(self, HTTP_REQUEST, 7, "Response: version=%s, response_code=%s, response_msg=%s", self->response_version, self->response, self->response_msg);
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

#undef SKIP_SPACES
#undef COPY_SPACE

static gboolean
http_parse_version(HttpProxy *self, gchar *version)
{
  z_func_enter(self, HTTP_DEBUG);
  if (strcasecmp(version, "HTTP/1.1") == 0)
    {
      self->proto_version = 0x0101;
    }
  else if (strcasecmp(version, "HTTP/1.0") == 0)
    {
      self->proto_version = 0x0100;
    }
  else if (version[0] == 0)
    {
      self->proto_version = 0x0009;
    }
  else
    {
      /* unknown protocol version */
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static void
http_insert_headers(gchar *key, ZPolicyObj *f, GString *headers)
{
  guint filter_type;
  gchar *value;

  if (!http_hash_get_type(f, &filter_type))
    {
      /* filter has no type field */
      return;
    }
  switch (filter_type)
    {
    case HTTP_INSERT:
    case HTTP_REPLACE:
      if (!z_policy_var_parse(f, "(is)", &filter_type, &value))
	{
	  /* error parsing HTTP_INSERT or HTTP_REPLACE rule */
	  return;
	}
      g_string_append(headers, key);
      g_string_append(headers, ": ");
      g_string_append(headers, value);
      g_string_append(headers, "\r\n");
      break;
    default:
      break;
    }
}

static gboolean
http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
{
  GString *headers = self->headers[side];
  GString *new_headers = g_string_sized_new(headers->len);
  GHashTable *hash = (side == EP_CLIENT) ? self->request_header_policy : self->response_header_policy;
  gchar *src = headers->str, *eol, *colon;
  gint action;

  z_func_enter(self, HTTP_DEBUG);  
  while (*src)
    {
      eol = strchr(src, '\n');
      colon = strchr(src, ':');
      if (eol && colon && (colon < eol))
	{
	  gchar *name, *value;
	  guint name_len, value_len;
	  gboolean dont_append = FALSE;
	  ZPolicyObj *f;

	  eol--;
	  if (*eol != '\r')
	    {
	      /* header not terminated by CRLF */
	      g_string_free(new_headers, 1);
	      z_func_leave(self, HTTP_DEBUG);
	      return FALSE;
	    }
	  eol--;

	  name = src;
	  name_len = colon - src;

	  colon++;
	  while (*colon == ' ') colon++;

	  value = colon;
	  value_len = eol - colon + 1;
	  src = eol + 3;
	  
	  g_string_truncate(self->current_header_name, 0);
	  g_string_append_len(self->current_header_name, name, name_len);
	  g_string_truncate(self->current_header_value, 0);
	  g_string_append_len(self->current_header_value, value, value_len);

	  if (filter)
	    action = filter(self, self->current_header_name, self->current_header_value);
	  else
	    action = -1;

	  f = g_hash_table_lookup(hash, self->current_header_name->str);
	  if (!f)
	    f = g_hash_table_lookup(hash, "*");
	  if (f)
	    {
	      guint filter_type;
	      ZPolicyObj *handler, *res;

	      z_policy_lock(self->super.thread);
	      if (!http_hash_get_type(f, &filter_type))
		{
		  /* filter has no type field */
		  z_policy_unlock(self->super.thread);
		  z_func_leave(self, HTTP_DEBUG);
		  return FALSE;
		}
	      z_policy_unlock(self->super.thread);
	      switch (filter_type)
		{
		case HTTP_POLICY_CALL:
		  z_policy_lock(self->super.thread);
		  if (!z_policy_var_parse(f, "(iO)", &filter_type, &handler))
		    {
		      /* error parsing HTTP_POLICY_CALL rule */
		      z_func_leave(self, HTTP_DEBUG);
		      return FALSE;
		    }
		  res = z_policy_call_object(handler, 
					     z_policy_var_build("(s#s#)", 
								name, name_len, 
								value, value_len));
		  if (!z_policy_var_parse(res, "i", &action))
		    {
		      z_proxy_log(self, HTTP_POLICY, 1, "policy callback for header [%s] returned non-int value", self->current_header_name->str);
		      z_policy_var_unref(res);
		      z_policy_unlock(self->super.thread);
		      z_func_leave(self, HTTP_DEBUG);
		      return FALSE;
		    }
		  z_policy_var_unref(res);
		  z_policy_unlock(self->super.thread);
		  break;
		case HTTP_PASS:
		case HTTP_INSERT:
		  break;
		case HTTP_REPLACE:
		case HTTP_DROP:
		  dont_append = TRUE;
		  break;
		case HTTP_CHANGE_NAME:
		  z_policy_lock(self->super.thread);
		  if (!z_policy_var_parse(f, "(is)", &filter_type, &name))
		    {
		      /* invalid CHANGE_NAME rule */
		      z_policy_unlock(self->super.thread);
		      z_func_leave(self, HTTP_DEBUG);
		      return FALSE;
		    }
		  g_string_assign(self->current_header_name, name);
		  z_policy_unlock(self->super.thread);
		  break;
		case HTTP_CHANGE_VALUE:
		  z_policy_lock(self->super.thread);
		  if (!z_policy_var_parse(f, "(is)", &filter_type, &value))
		    {
		      /* invalid CHANGE_VALUE rule */
		      z_policy_unlock(self->super.thread);
		      z_func_leave(self, HTTP_DEBUG);
		      return FALSE;
		    }
		  g_string_assign(self->current_header_value, value);
		  z_policy_unlock(self->super.thread);
		  break;
		case HTTP_CHANGE_BOTH:
		  z_policy_lock(self->super.thread);
		  if (!z_policy_var_parse(f, "(iss)", &filter_type, &name, &value))
		    {
		      /* invalid CHANGE_BOTH rule */
		      z_policy_unlock(self->super.thread);
		      z_func_leave(self, HTTP_DEBUG);
		      return FALSE;
		    }
		  g_string_assign(self->current_header_name, name);
		  g_string_assign(self->current_header_value, value);
		  z_policy_unlock(self->super.thread);
		  break;
		}
	    }
	  switch (action)
	    {
	    case Z_UNSPEC:
	    case Z_ACCEPT:
	      break;
	    case Z_REJECT:
	    case Z_ABORT:
	      self->error_code = HTTP_MSG_POLICY_VIOLATION;
	      z_func_leave(self, HTTP_DEBUG);
	      return FALSE;
	    case Z_DENY:
	    case Z_DROP:
	      dont_append = 1;
	      break;
	    }
	  if (!dont_append)
	    {
	      g_string_append(new_headers, self->current_header_name->str);
	      g_string_append(new_headers, ": ");
	      g_string_append(new_headers, self->current_header_value->str);
	      g_string_append(new_headers, "\r\n");
	    }
	}
      else
	{
	  if (*(eol - 1) == '\r' && (src == eol - 1))
	    {
	      /* empty line */
	      break;
	    }
	  else
	    {
	      g_string_free(new_headers, 1);
	      z_func_leave(self, HTTP_DEBUG);
	      return FALSE;
	    }
	}
    }
  g_string_assign(headers, new_headers->str);

  z_policy_lock(self->super.thread);
  g_hash_table_foreach(hash, (GHFunc) http_insert_headers, headers);
  z_policy_unlock(self->super.thread);

  g_string_append(headers, "\r\n");
  g_string_free(new_headers, 1);
  z_func_leave(self, HTTP_DEBUG);
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean 
http_data_transfer(HttpProxy *self, guint from, guint to, gboolean expect_data)
{
  GString *headers = self->headers[from];
  gint content_length;
  HttpTransferFilter filter_transfer;
  gchar *hdr;
  guint hdr_len;

  z_func_enter(self, HTTP_DEBUG);
  filter_transfer = http_transfer_filter_copy;

  content_length = -2;  // -1 unknown, -2 no data

  if (http_header_lookup(headers, "Transfer-Encoding", &hdr, &hdr_len, TRUE) &&
      strncasecmp(hdr, "chunked", MIN(hdr_len, 7)) == 0)
    {
      filter_transfer = http_transfer_filter_chunked;
      content_length = -1;
    }
  else if (http_header_lookup(headers, "Content-Length", &hdr, &hdr_len, TRUE))
    {
      char buf[hdr_len + 1];
      gchar *end;
      
      memcpy(buf, hdr, hdr_len);
      buf[hdr_len] = 0;
      content_length = strtoul(buf, &end, 10);
      if (end - buf != hdr_len)
	{
	  /* content-length not a number */
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      if (content_length == 0)
	content_length = -2;
      
      filter_transfer = http_transfer_filter_copy;
    }
  else if (expect_data)
    {
      content_length = -1;
      self->connection_mode = HTTP_CONNECTION_CLOSE; 
    }
  else 
    {
      content_length = -2;
    }

  if (content_length != -2)
    {
      if (!filter_transfer(self, from, to, content_length))
	{
	  /* FIXME: transfer unsuccessful */
	  z_proxy_log(self, HTTP_ERROR, 2, "Data transfer failed.");
	  self->error_code = HTTP_MSG_OK; /* error message may corrupt data, close connection silently */
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

/* FIXME: resolve continuation lines */
static gboolean
http_fetch_headers(HttpProxy *self, int side)
{
  GString *headers = self->headers[side];
  gchar *line;
  guint res, line_length;
  guint count = 0;

  z_func_enter(self, HTTP_DEBUG);
  /* check if we have to fetch request headers */
  if (self->proto_version < 0x0100)
    {
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    }
  
  g_string_assign(headers, "");
  while (1)
    {
      gchar *colon;
      
      res = z_read_line_get(self->readline[side], &line, &line_length);
      if (ST_ERROR(res) != ST_OK)
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      if (line_length == 0)
        {
          break;
        }
      for (colon = line; colon - line < line_length && *colon != ' ' && *colon != ':'; colon++)
        ;
      if (*colon != ':')
        {
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
        
      /* strip trailing white space */
      while (line_length > 0 && line[line_length - 1] == ' ') 
        line_length--;
        
      g_string_append_len(headers, line, line_length);
      
      g_string_append(headers, "\r\n");
      count++;
      if (count > self->max_header_lines)
        {
          /* too many headers */
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
    }
  g_string_append(headers, "\r\n");
  
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_fetch_request(HttpProxy *self)
{
  gchar *line;
  guint line_length;
  guint res;

  z_func_enter(self, HTTP_DEBUG); 
  while (1) 
    {
      res = z_read_line_get(self->readline[EP_CLIENT], &line, &line_length);
      if (ST_ERROR(res) != ST_OK)
        {
          if (errno == ETIME)
            {
              self->error_code = HTTP_MSG_OK;
            }
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      if (line_length != 0)
	break;
    }
  if (!http_split_request(self, line, line_length))
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  self->request_flags = http_proto_request_lookup(self->request_method);
  if (self->request_flags == -1)
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
    
  if (!http_parse_version(self, self->request_version))
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if (!http_fetch_headers(self, EP_CLIENT))
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}


static gboolean
http_process_request(HttpProxy *self)
{
  gchar *hdr;
  guint hdr_length;

  z_func_enter(self, HTTP_DEBUG);  
  if (self->proto_version > 0x0100)
    self->connection_mode = HTTP_CONNECTION_KEEPALIVE;
  else    
    self->connection_mode = HTTP_CONNECTION_CLOSE;
    
  if (self->request_flags & HTTP_REQ_FLG_CONNECT)
    {
      if (self->proto_version >= 0x0100)
        {
          self->request_type = HTTP_REQTYPE_PROXY;
          z_func_leave(self, HTTP_DEBUG);
          return TRUE;
        }
      else
        {
          self->error_code = HTTP_MSG_CLIENT_SYNTAX;
          g_string_sprintf(self->error_info, "CONNECT method requires HTTP/1.0 or later");
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
    }
    
  if (http_header_lookup(self->headers[EP_CLIENT], "Proxy-Connection", &hdr, &hdr_length, 1))
    {
      if (http_header_lookup(self->headers[EP_CLIENT], "Connection", &hdr, &hdr_length, 1))
	{
	  /* both Proxy-Connection & Connection headers ... */
	  self->error_code = HTTP_MSG_CLIENT_SYNTAX;
	  g_string_sprintf(self->error_info, "Both Proxy-Connection and Connection headers exist.");
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      self->request_type = HTTP_REQTYPE_PROXY;
    }
  else if (http_header_lookup(self->headers[EP_CLIENT], "Connection", &hdr, &hdr_length, 1))
    {
      self->request_type = HTTP_REQTYPE_SERVER;
    }
  else if (self->request_url->str[0] != '/')
    {
      /* neither Connection nor Proxy-Connection header exists, and URI
         doesn't seem to be a simple filename */
      self->request_type = HTTP_REQTYPE_PROXY;
    }
  else
    {
      /* default */
      self->request_type = HTTP_REQTYPE_SERVER;
    }

  if (self->transparent_mode &&
      ((self->request_type == HTTP_REQTYPE_PROXY && !self->transparent_proxy_requests) ||
       (self->request_type == HTTP_REQTYPE_SERVER && !self->transparent_server_requests)))
    {
      /* */
      self->error_code = HTTP_MSG_POLICY_VIOLATION;
      g_string_sprintf(self->error_info, "%s requests not permitted in transparent mode.", 
		       self->request_type == HTTP_REQTYPE_SERVER ? "server" : "proxy");
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if (((self->request_type == HTTP_REQTYPE_SERVER) && 
       http_header_lookup(self->headers[EP_CLIENT], "Connection", &hdr, &hdr_length, 1)) ||
      ((self->request_type == HTTP_REQTYPE_PROXY) && 
       http_header_lookup(self->headers[EP_CLIENT], "Proxy-Connection", &hdr, &hdr_length, 1)))
    {
      if (strncasecmp(hdr, "Keep-Alive", hdr_length) == 0)
        {
          self->connection_mode = HTTP_CONNECTION_KEEPALIVE;
        }
      else if (strncasecmp(hdr, "Close", hdr_length) == 0)
        {
          self->connection_mode = HTTP_CONNECTION_CLOSE;
        }
    }

  if (!http_check_url(self->request_url->str))
    {
      /* invalid url */
      self->error_code = HTTP_MSG_INVALID_URL;
      g_string_sprintf(self->error_info, "Invalid url %s", self->request_url->str);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if (http_header_lookup(self->headers[EP_CLIENT], "Host", &hdr, &hdr_length, 1))
    {
      int len = MIN(hdr_length, sizeof(self->remote_server) - 1);

      memcpy(self->remote_server, hdr, len);
      self->remote_server[len] = 0;
    }

  if (self->transparent_mode)
    {
      if (strncasecmp(self->request_url->str, "http://", 7) != 0 &&
          strncasecmp(self->request_url->str, "ftp://", 6) != 0)
	{
	  gchar buf[hdr_length + 12];
	  
	  /* no protocol description */
	  if (!self->remote_server[0])
	    {
	      /* no host header */
	      /* FIXME: make this case configurable */
	      z_proxy_log(self, HTTP_REQUEST, 6, "No host header in request, using numeric IP");
	      strcpy(self->remote_server, "unknown");
	    }
	  g_snprintf(buf, sizeof(buf), "http://%s", self->remote_server);
	  g_string_prepend(self->request_url, buf);
	}
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;  
} 

static guint 
http_request_filter_connection_header(HttpProxy *self, GString *name, GString *value)
{
  z_func_enter(self, HTTP_DEBUG);
  if (self->rewrite_host_header &&
      strcasecmp(name->str, "Host") == 0)
    {
      /* NOTE: the constant 80 below is intentional, if
       * default_port is changed we still have to send
       * correct host header (containing port number)
       */
      if (self->remote_port != 80)
	g_string_sprintf(value, "%s:%d", self->remote_server, self->remote_port);
      else
	g_string_assign(value, self->remote_server);
      
      z_func_leave(self, HTTP_DEBUG);
      return Z_ACCEPT;
    }
  switch (self->request_type)
    {
    case HTTP_REQTYPE_SERVER:
      /* if we have a parent proxy, Connection -> Proxy-Connection
       * otherwise leave it as is 
       */
      if (self->parent_proxy->len &&
	  strcasecmp(name->str, "Connection") == 0)
	g_string_assign(name, "Proxy-Connection");
      break;
    case HTTP_REQTYPE_PROXY:
      /* if we have a parent proxy leave it as is
       * otherwise Proxy-Connection -> Connection 
       */
      if (self->parent_proxy->len == 0 &&
	  strcasecmp(name->str, "Proxy-Connection") == 0)
	g_string_assign(name, "Connection");
      break;
    }
  z_func_leave(self, HTTP_DEBUG);
  return Z_ACCEPT;
}

static gboolean
http_filter_request(HttpProxy *self)
{
  ZPolicyObj *f;
  gint rc;

  z_func_enter(self, HTTP_DEBUG);
  f = g_hash_table_lookup(self->request_method_policy, self->request_method);
  if (!f)
    f = g_hash_table_lookup(self->request_method_policy, "*");
  if (f)
    {
      ZPolicyObj *handler, *res;
      gint filter_type;
      
      z_policy_lock(self->super.thread);  
      if (!http_hash_get_type(f, &filter_type))
        {
	  z_proxy_log(self, HTTP_POLICY, 1, "invalid request hash item(%s)", self->request_method);
	  z_policy_unlock(self->super.thread);
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
          
        }
      z_policy_unlock(self->super.thread);

      switch (filter_type)
        {
        case HTTP_POLICY_CALL:
	  z_policy_lock(self->super.thread);
          if (!z_policy_var_parse(f, "(iO)", &filter_type, &handler))
            {
              z_proxy_log(self, HTTP_POLICY, 1, "error parsing HTTP_POLICY_CALL rule");
              z_policy_unlock(self->super.thread);
              z_func_leave(self, HTTP_DEBUG);
              return FALSE;
            }
          res = z_policy_call_object(handler, z_policy_var_build("(sss)", self->request_method, self->request_url->str, self->request_version));
          if (!z_policy_var_parse(res, "i", &rc))
            {
              rc = Z_DROP;
            }
          z_policy_var_unref(res);
	  z_policy_unlock(self->super.thread);
          break;
	case HTTP_PASS:
	  rc = Z_ACCEPT;
	  break;
	case HTTP_DROP:
	  /* dropped command */
	  rc = Z_DROP;
	  break;
	default:
	  z_proxy_log(self, HTTP_POLICY, 1, "invalid request hash item, request=%s", self->request_method);
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
        }
      switch (rc)
        {
        case Z_ACCEPT:
          break;
        case Z_DENY:
        case Z_REJECT:
        case Z_DROP:
	  self->error_code = HTTP_MSG_POLICY_VIOLATION;
	  g_string_sprintf(self->error_info, "Method %s denied by policy", self->request_method);
	  z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
      if (self->proto_version >= 0x0100)
        {
          z_proxy_log(self, HTTP_REQUEST, 7, "prefilter request=%s", self->headers[EP_CLIENT]->str);
          if (!http_filter_headers(self, EP_CLIENT, http_request_filter_connection_header))
            {
              z_func_leave(self, HTTP_DEBUG);
              return FALSE;
            }
          z_proxy_log(self, HTTP_REQUEST, 7, "postfilter request=%s", self->headers[EP_CLIENT]->str);
        }
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    }
  self->error_code = HTTP_MSG_POLICY_VIOLATION;
  g_string_sprintf(self->error_info, "Method %s denied by policy", self->request_method);
  z_func_leave(self, HTTP_DEBUG);
  return FALSE;
}

static gboolean
http_connect_server(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  if (!self->super.endpoints[EP_SERVER] ||
      strcasecmp(self->remote_server, self->connected_server) != 0 ||
      self->remote_port != self->connected_port)
    {
      if (self->super.endpoints[EP_SERVER])
	{
	  z_read_line_destroy(self->readline[EP_SERVER]);
	  z_stream_shutdown(self->super.endpoints[EP_SERVER], SHUT_RDWR);
	  z_stream_close(self->super.endpoints[EP_SERVER]);
	  self->readline[EP_SERVER] = NULL;
	  self->super.endpoints[EP_SERVER] = NULL;
	}
      if (!z_proxy_connect_server_event(&self->super, self->remote_server, self->remote_port))
	{
	  /* error connecting to server */
	  self->error_code = HTTP_MSG_CONNECT_ERROR;
	  self->error_response = 502;
	  g_string_sprintf(self->error_info, "Error establishing connection to %s", self->remote_server);
	  z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}

      strcpy(self->connected_server, self->remote_server);
      self->connected_port = self->remote_port;
      if (!http_server_stream_init(self))
        {
          /* should never happen */
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_copy_request(HttpProxy *self)
{
  gchar *request;
  gchar *proto, *userpw, *host, *file;
  guint proto_len, userpw_len, host_len, file_len;
  int attempts = 0;
  
  z_func_enter(self, HTTP_DEBUG);
  if (!http_split_url(self->request_url->str, 
		      &proto, &proto_len, 
		      &userpw, &userpw_len,
		      &host, &host_len,
		      &self->remote_port,
		      &file, &file_len))
    {
      /* invalid URL */
      self->error_code = HTTP_MSG_INVALID_URL;
      g_string_sprintf(self->error_info, "Badly formatted url: %s", self->request_url->str);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  if (self->remote_port == 0)
    self->remote_port = self->default_port;

  if (strncasecmp(proto, "http://", 7) &&
      strncasecmp(proto, "ftp://", 6))
    {
      /* unsupported protocol */
      self->error_code = HTTP_MSG_CLIENT_SYNTAX;
      g_string_sprintf(self->error_info, "Unsupported protocol: %.*s", (int) proto_len, proto);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if (self->parent_proxy->len)
    {
      host_len = MIN(self->parent_proxy->len, sizeof(self->remote_server) - 1);
      memcpy(self->remote_server, self->parent_proxy->str, host_len);
      self->remote_port = self->parent_proxy_port;
    }
  else
    {
      host_len = MIN(host_len, sizeof(self->remote_server) - 1);
      memcpy(self->remote_server, host, host_len);
    }

  self->remote_server[host_len] = 0;

  while (attempts < 4)
    {
      attempts++;
      if (!http_connect_server(self))
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      if (self->proto_version >= 0x0100)
	{
	  request = alloca(strlen(self->request_method) + self->request_url->len + strlen(self->request_version) + self->headers[EP_CLIENT]->len + 8);
	  if (self->parent_proxy->len)
	    sprintf(request, "%s %s %s\r\n%s", self->request_method, self->request_url->str, self->request_version, self->headers[EP_CLIENT]->str);
	  else
	    sprintf(request, "%s %.*s %s\r\n%s", self->request_method, (int) file_len, file, self->request_version, self->headers[EP_CLIENT]->str);
	}
      else
	{
	  request = alloca(strlen(self->request_method) + self->request_url->len + 5);
	  sprintf(request, "%s %s\r\n", self->request_method, self->request_url->str);
	}
  
      if (!http_write(self, EP_SERVER, request, strlen(request)))
        {
          if (attempts < 4)
	    continue;
	  
	  self->error_code = HTTP_MSG_CONNECT_ERROR;
	  self->error_response = 502;
	  g_string_sprintf(self->error_info, "Error establishing connection to %s", self->remote_server);

          z_func_leave(self, HTTP_DEBUG)
          return FALSE;
	}
    
      if (http_data_transfer(self, EP_CLIENT, EP_SERVER, FALSE))
	break;
      else 
        {
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }

    }

  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean 
http_fetch_response(HttpProxy *self)
{
  gchar *line;
  guint line_length;
  guint res;
  gboolean reply09;

  z_func_enter(self, HTTP_DEBUG);
  self->response[0] = 0;
  self->response_code = -1;
  if (self->proto_version < 0x0100)
    {
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    }
  
  while (self->response_code == -1 || self->response_code == 100)
    {
      res = z_read_line_get(self->readline[EP_SERVER], &line, &line_length);
      if (ST_ERROR(res) != ST_OK)
        {
          self->error_code = HTTP_MSG_OK;
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      
      if (!http_split_response(self, line, line_length, &reply09))
        {
          if (reply09)
            {
              z_read_line_unget(self->readline[EP_SERVER], line, line_length);
	      self->proto_version = 0x0009;
	      z_func_leave(self, HTTP_DEBUG);
	      return TRUE;
            }
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      
      self->response_flags = http_proto_response_lookup(self->response);
      if (self->response_flags == -1)
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      if (!http_parse_version(self, self->response_version))
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
      
      if (!http_fetch_headers(self, EP_SERVER))
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
    }

  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_process_response(HttpProxy *self)
{
  gchar *hdr;
  guint hdr_length;

  z_func_enter(self, HTTP_DEBUG);
  /* previously set by http_parse_version */
  switch (self->proto_version)
    {
    case 0x0009:
      strncpy(self->response, "200", sizeof(self->response_code));
      self->response_code = 200;
      self->response_flags = HTTP_RESP_FLG_TRANSFER | HTTP_RESP_FLG_DATA_COND;
      self->connection_mode = HTTP_CONNECTION_CLOSE;
      z_func_leave(self, HTTP_DEBUG);
      return TRUE;
    case 0x0100:
      self->connection_mode = HTTP_CONNECTION_CLOSE;
      break;
    case 0x0101:
      /* leave connection mode as is, as was requested by
	 the browser */
      break;
    default:
      z_proxy_log(self, HTTP_ERROR, 1, "Unsupported protocol version 0x%04x", self->proto_version);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  if ((self->request_type == HTTP_REQTYPE_SERVER && 
       http_header_lookup(self->headers[EP_SERVER], "Connection", &hdr, &hdr_length, 1)) ||
      (self->request_type == HTTP_REQTYPE_PROXY && 
       http_header_lookup(self->headers[EP_SERVER], "Proxy-Connection", &hdr, &hdr_length, 1)))
    {
      if (strncasecmp(hdr, "Keep-Alive", hdr_length) == 0)
        {
          self->connection_mode = HTTP_CONNECTION_KEEPALIVE;
        }
      else 
        {
          self->connection_mode = HTTP_CONNECTION_CLOSE;
        }
    }

  if (self->response_flags & HTTP_RESP_FLG_STOP)
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;  /* FIXME: not implemented */
    }

  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static guint
http_response_filter_connection_header(HttpProxy *self, GString *name, GString *value)
{
  z_func_enter(self, HTTP_DEBUG);
  switch (self->request_type)
    {
    case HTTP_REQTYPE_SERVER:
      /* if we have a parent proxy, Proxy-Connection -> Connection
       * otherwise leave it as is 
       */
      if (self->parent_proxy->len &&
	  strcasecmp(name->str, "Proxy-Connection") == 0)
	g_string_assign(name, "Connection");
      if (self->connection_mode == HTTP_CONNECTION_CLOSE &&
	  strcasecmp(name->str, "Connection") == 0)
	{
	  g_string_assign(value, "close");
	}
      break;
    case HTTP_REQTYPE_PROXY:
      /* if we have a parent proxy leave it as is
       * otherwise Connection -> Proxy-Connection 
       */
      if (self->parent_proxy->len == 0 &&
	  strcasecmp(name->str, "Connection") == 0)
	g_string_assign(name, "Proxy-Connection");

      if (self->connection_mode == HTTP_CONNECTION_CLOSE &&
	  strcasecmp(name->str, "Proxy-Connection") == 0)
	{
	  g_string_assign(value, "close");
	}
      break;
    }
  z_func_leave(self, HTTP_DEBUG);
  return Z_ACCEPT;
}

static gboolean
http_filter_response(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  if (self->proto_version >= 0x0100)
    {
      z_proxy_log(self, HTTP_RESPONSE, 7, "prefilter response=%s", self->headers[EP_SERVER]->str);
      if (!http_filter_headers(self, EP_SERVER, http_response_filter_connection_header))
        {
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
      z_proxy_log(self, HTTP_RESPONSE, 7, "postfilter response=%s", self->headers[EP_SERVER]->str);
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_copy_response(HttpProxy *self)
{
  z_func_enter(self, HTTP_DEBUG);
  if (self->proto_version >= 0x0100)
    {
      gchar *response; 
      
      response = alloca(strlen(self->response_version) + strlen(self->response) + strlen(self->response_msg) + self->headers[EP_SERVER]->len + 6);
      sprintf(response, "%s %s %s\r\n%s", self->response_version, self->response, self->response_msg, self->headers[EP_SERVER]->str);
      if (!http_write(self, EP_CLIENT, response, strlen(response)))
        {
          z_func_leave(self, HTTP_DEBUG);
          return FALSE;
        }
    }

  if ((((self->request_flags & HTTP_REQ_FLG_HEAD) != 0) && 
       ((self->response_flags & HTTP_RESP_FLG_DATA_COND) != 0)) ||       //head & data_cond OR
      (((self->response_flags & HTTP_RESP_FLG_DATA) == 0) && 
       ((self->response_flags & HTTP_RESP_FLG_DATA_COND) == 0)))          //!data & !data_cond
    {
      /* no need to transfer data */
      
    }
  else
    {
      if (!http_data_transfer(self, EP_SERVER, EP_CLIENT, TRUE))
        {
          z_func_leave(self, HTTP_DEBUG);
	  return FALSE;
	}
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static gboolean
http_handle_connect(HttpProxy *self)
{
  /* connect is enabled here */
  int i, rc; 
  ZPolicyObj *res;
  gchar *colon, *end, *success;
  gboolean called;

  z_func_enter(self, HTTP_DEBUG);
  if (self->parent_proxy->len)
    {
      z_proxy_log(self, HTTP_POLICY, 1, "CONNECT method is not supported with parent proxies");
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }

  colon = strchr(self->request_url->str, ':');
  if (colon)
    {
      *colon = 0;
      strncpy(self->remote_server, self->request_url->str, sizeof(self->remote_server) - 1);
      self->remote_server[sizeof(self->remote_server)] = 0;
      *colon = ':';
      self->remote_port = strtoul(colon + 1, &end, 10);
    }
  else
    {
      self->error_code = HTTP_MSG_CLIENT_SYNTAX;
      g_string_sprintf(self->error_info, "Invalid CONNECT request.");
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
          
  if (!http_connect_server(self))
    {
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  
  z_proxy_log(self, HTTP_DEBUG, 6, "starting connect method");
  
  z_policy_lock(self->super.thread);
  res = z_policy_call(self->super.handler, "connectMethod", z_policy_var_build("()"), &called);
  
  if (!z_policy_var_parse(res, "i", &rc))
    {
      self->error_code = HTTP_MSG_POLICY_SYNTAX;
      z_policy_var_unref(res);
      z_policy_unlock(self->super.thread);
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  z_policy_var_unref(res);
  z_policy_unlock(self->super.thread);
  
  success = "HTTP/1.0 200 Connection established\r\n\r\n";
  if (!http_write(self, EP_CLIENT, success, strlen(success)))
    {
      /* hmm... I/O error */
      z_func_leave(self, HTTP_DEBUG);
      return FALSE;
    }
  
  /* release, but don't close fds */
  for (i = EP_CLIENT; i < EP_MAX; i++)
    {
      z_read_line_destroy(self->readline[i]);
      self->readline[i] = NULL;
      z_stream_unref(self->super.endpoints[i]);
      self->super.endpoints[i] = NULL;
    }
  z_func_leave(self, HTTP_DEBUG);
  return TRUE;
}

static void
http_main(HttpProxy *self)
{
  int requests = 0;
  
  z_func_enter(self, HTTP_DEBUG);
  /* loop for keep-alive */
  while (1)
    {
      self->error_code = -1;
      z_proxy_log(self, HTTP_DEBUG, 6, "fetching request and headers");
      /* fetch request */
      if (!http_fetch_request(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_CLIENT_SYNTAX;
	  break;
	}
      z_proxy_log(self, HTTP_DEBUG, 6, "processing request and headers");
      if (!http_process_request(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_CLIENT_SYNTAX;
	  break;
	}
      z_proxy_log(self, HTTP_DEBUG, 6, "filtering request and headers");
      if (!http_filter_request(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_POLICY_SYNTAX;
	  break;
	}
      
      if ((self->request_flags & HTTP_REQ_FLG_CONNECT))
        {
          if (http_handle_connect(self))
            {
              /* connect method was successful, we can now safely quit */
              z_func_leave(self, HTTP_DEBUG);
              return;
            }
          else
            break;
        }
      if (self->max_keepalive_requests != 0 && 
          requests > self->max_keepalive_requests)
        {
          z_proxy_log(self, HTTP_DEBUG, 6, "max_keepalive_request reached, setting connection mode to close");
          self->connection_mode = HTTP_CONNECTION_CLOSE;
        }
      z_proxy_log(self, HTTP_DEBUG, 6, "sending request and headers, copying request data");
      if (!http_copy_request(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_CLIENT_SYNTAX;
	  break;
	}
      z_proxy_log(self, HTTP_DEBUG, 6, "fetching response and headers");
      if (!http_fetch_response(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_SERVER_SYNTAX;
	  break;
	}
      z_proxy_log(self, HTTP_DEBUG, 6, "processing response and headers");
      if (!http_process_response(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_SERVER_SYNTAX;
          break;
        }
      z_proxy_log(self, HTTP_DEBUG, 6, "filtering response and headers");
      if (!http_filter_response(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_POLICY_SYNTAX;
          break;
        }
      z_proxy_log(self, HTTP_DEBUG, 6, "copying response and headers, copying response data");
      if (!http_copy_response(self))
        {
          if (self->error_code == -1)
            self->error_code = HTTP_MSG_SERVER_SYNTAX;
          break;
        }
      if (self->connection_mode == HTTP_CONNECTION_CLOSE ||
	  self->response_flags & HTTP_RESP_FLG_CLOSE)
	break;
	
      requests++;
    }
  z_proxy_log(self, HTTP_DEBUG, 6, "exiting keep-alive loop");
  if (self->error_code != -1)
    {
      http_error_message(self, self->error_response, self->error_code, self->error_info);
    }
  z_func_leave(self, HTTP_DEBUG);
}

static void
http_thread(void *s)
{
  HttpProxy *self = (HttpProxy *) s;

  z_func_enter(self, HTTP_DEBUG);
  http_config_set_defaults(self);
  http_register_vars(self);

  ZPROXY_SET_STATE(self, PS_CONFIG);
  z_proxy_config_event(&self->super);
  
  http_config_init(self);
  http_client_stream_init(self);

  ZPROXY_SET_STATE(self, PS_STARTING_UP);
  z_proxy_startup_event(&self->super);

  ZPROXY_SET_STATE(self, PS_WORKING);

  http_main(self);

  ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event(&self->super);

  z_proxy_destroy(&self->super);

  z_func_leave(self, HTTP_DEBUG);
}

static void
http_proxy_free(ZProxy *s)
{
  HttpProxy *self = (HttpProxy *) s;
  int i;

  z_func_enter(self, HTTP_DEBUG);  
  for (i = EP_CLIENT; i < EP_MAX; i++)
    {
      if (self->readline[i])
        {
          z_read_line_destroy(self->readline[i]);
          self->readline[i] = NULL;
        }
      g_string_free(self->headers[i], TRUE);
    }

  g_string_free(self->current_header_name, TRUE);
  g_string_free(self->current_header_value, TRUE);
  g_string_free(self->request_url, TRUE);

  g_string_free(self->parent_proxy, TRUE);
  g_string_free(self->error_info, TRUE);
  /* NOTE: hashes are freed up by pyvars */
  z_func_leave(self, HTTP_DEBUG);
}

static ZProxy *
http_proxy_new(gchar *session_id, ZStream *client, ZPolicyObj *handler)
{
  HttpProxy *self = g_new0(HttpProxy, 1);

  z_proxy_init(&self->super, session_id, client, handler);
  self->super.free = http_proxy_free;
  
  z_thread_new(session_id, http_thread, self);
  return (&self->super);
}

gint
zorp_module_init(void)
{
  z_registry_add("http", ZR_PROXY, http_proxy_new);
  return TRUE;
}
