/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: plug.c,v 1.28 2001/01/15 13:47:02 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/stream.h>
#include <zorp/proxy.h>
#include <zorp/poll.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/log.h>
#include <zorp/registry.h>
#include <zorp/sockaddr.h>
#include <zorp/io.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

#define ENABLE_PKT_STATS 1

#define PROXY_BUFSIZE    1500 /* 64k */

typedef struct _ZIOBuffer 
{
  gchar buf[PROXY_BUFSIZE];
  guint ofs, end;
  guint packet_count, packet_bytes;
} ZIOBuffer;

typedef struct _PlugProxy
{
  ZProxy super;
  ZStream *downstreams[EP_MAX];
  ZIOBuffer buffers[EP_MAX];
  ZIOBuffer *downbufs[EP_MAX];
  ZPoll *poll;
  gint timeout;
  gint packet_stats_interval;
  guint global_packet_count;
  gboolean copy_to_server, copy_to_client;
  struct timeval started_time;
} PlugProxy;

#if 0
static char *ep_names[] = 
{ "client", "server" };
#endif

static void plug_proxy_free(ZProxy *s);

static guint
plug_read_input(PlugProxy *self, ZStream *input, ZIOBuffer *buf)
{
  gint rc;
  gboolean called;

  rc = z_stream_read(input, buf->buf, sizeof(buf->buf), &buf->end);
  switch (rc)
    {
    case G_IO_ERROR_NONE:

      buf->packet_bytes += buf->end;
      buf->packet_count++;
      self->global_packet_count++;
      if (self->packet_stats_interval &&
	  (self->global_packet_count % self->packet_stats_interval) == 0)
	{
	  int i;
	  ZPolicyObj *res;

	  z_policy_lock(self->super.thread);
	  for (i = EP_CLIENT; i < EP_MAX; i++)
	    {
	      ZIOBuffer *buf;
              struct timeval now, spent;
	      char *varname[EP_MAX] = { "bandwidth_to_client", "bandwidth_to_server" };

	      buf = &self->buffers[i];
      
              gettimeofday(&now, NULL);
              spent.tv_sec = now.tv_sec - self->started_time.tv_sec;
              spent.tv_usec = now.tv_usec - self->started_time.tv_usec;
              if (spent.tv_usec < -500000)
	        {
	          spent.tv_sec++;
		}
	      z_policy_setattr(self->super.handler, varname[i], z_policy_var_build("d", ((double) buf->packet_bytes) / spent.tv_sec));
	    }
	  res = z_policy_call(self->super.handler, "packetStats", 
	  		      z_policy_var_build("iiii", 
					         self->buffers[EP_CLIENT].packet_bytes, 
					         self->buffers[EP_CLIENT].packet_count, 
					         self->buffers[EP_SERVER].packet_bytes, 
					         self->buffers[EP_SERVER].packet_count),
			      &called);
	  if (res)
	    {
              if (!z_policy_var_parse(res, "i", &rc))
                rc = Z_DENY;
            }
          else
            if (called)
              rc = Z_DENY;
          z_policy_var_unref(res);
	  z_policy_unlock(self->super.thread);
	  if (rc != Z_ACCEPT)
	    {
	      return ST_CLOSE | ST_EOF;
	    }
	}
      if (buf->end == 0)
        {
          return ST_CLOSE | ST_EOF;
        }
      return ST_OK | ST_GOON;
    case G_IO_ERROR_AGAIN:
      return ST_AGAIN | ST_GOON;
    default:
      return ST_ERR | ST_CLOSE;
    }
}

static guint
plug_write_output(PlugProxy *self, ZIOBuffer *buf, ZStream *output)
{
  int rc;
  guint bytes_written;
  
  if (buf->ofs != buf->end)
    {
      /* buffer not empty */
      rc = z_stream_write(output, &buf->buf[buf->ofs], buf->end - buf->ofs, &bytes_written);
      switch (rc) 
        {
        case G_IO_ERROR_NONE:
          buf->ofs += bytes_written;
          break;
        case G_IO_ERROR_AGAIN:
          break;
        default:
          return ST_ERR | ST_CLOSE;
        }
      if (buf->ofs != buf->end)
        {
          output->want_write = TRUE;
          return ST_AGAIN | ST_GOON;
        }
      else
        return ST_OK | ST_GOON;
    }
  return ST_OK | ST_GOON;
}

static guint
plug_copy_data(PlugProxy *self, ZStream *from, ZStream *to, ZIOBuffer *buf)
{
  int rc;
  int pkt_count = 0;

  from->want_read = FALSE;

  if (to)
    {
      to->want_write = FALSE;
      rc = plug_write_output(self, buf, to);
      if (ST_ERROR(rc) != ST_OK)
	return rc;
    }

  while (1) 
    {  
      buf->ofs = buf->end = 0;
      rc = plug_read_input(self, from, buf);
      if (ST_ERROR(rc) == ST_OK)
        {
	  if (to)
	    {
	      rc = plug_write_output(self, buf, to);
	      if (ST_ERROR(rc) == ST_AGAIN)
		break;
              else if (ST_FAILED(rc))
                {
                  z_message(0, "%s: Error writing %s (%m)", self->super.session_id, to->name);
                  return ST_QUIT | ST_OK;
                }
	    }
        }
      else if (ST_ERROR(rc) == ST_AGAIN)
        break;
      else if (ST_ERROR(rc) == ST_EOF)
        return ST_QUIT | ST_OK;
      else if (ST_FAILED(rc))
        {
          z_message(0, "%s: Error reading %s (%s)", self->super.session_id, from->name, g_strerror(errno));
          return ST_QUIT | ST_OK;
        }
      pkt_count++;
    }

  if (buf->ofs == buf->end)
    {
      from->want_read = TRUE;
    }
  return ST_OK | ST_GOON;
}

/* callbacks when no stacking is made */
static guint
plug_copy_client_to_server(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  if (self->copy_to_server)
    return plug_copy_data(self, self->super.endpoints[EP_CLIENT], self->super.endpoints[EP_SERVER], &self->buffers[EP_SERVER]);
  else
    return plug_copy_data(self, self->super.endpoints[EP_CLIENT], NULL, &self->buffers[EP_SERVER]);
}

static guint
plug_copy_server_to_client(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  if (self->copy_to_client)
    return plug_copy_data(self, self->super.endpoints[EP_SERVER], self->super.endpoints[EP_CLIENT], &self->buffers[EP_CLIENT]);
  else
    return plug_copy_data(self, self->super.endpoints[EP_SERVER], NULL, &self->buffers[EP_CLIENT]);
}

/* callbacks when a stacked module exists */
static guint
plug_copy_client_to_down(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  if (self->copy_to_server)
    return plug_copy_data(self, self->super.endpoints[EP_CLIENT], self->downstreams[EP_CLIENT], self->downbufs[EP_CLIENT]);
  else
    return plug_copy_data(self, self->super.endpoints[EP_CLIENT], NULL, self->downbufs[EP_CLIENT]);
}

static guint
plug_copy_down_to_client(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  return plug_copy_data(self, self->downstreams[EP_CLIENT], self->super.endpoints[EP_CLIENT], &self->buffers[EP_CLIENT]);

}

static guint
plug_copy_server_to_down(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  if (self->copy_to_client)
    return plug_copy_data(self, self->super.endpoints[EP_SERVER], self->downstreams[EP_SERVER], self->downbufs[EP_SERVER]);
  else
    return plug_copy_data(self, self->super.endpoints[EP_SERVER], NULL, self->downbufs[EP_SERVER]);
}

static guint
plug_copy_down_to_server(ZStream *stream, GIOCondition cond)
{
  PlugProxy *self = (PlugProxy *) stream->user_data;

  return plug_copy_data(self, self->downstreams[EP_SERVER], self->super.endpoints[EP_SERVER], &self->buffers[EP_CLIENT]);
}

static void
plug_init_streams(PlugProxy *self)
{
  if (!self->super.endpoints[EP_CLIENT] ||
      !self->super.endpoints[EP_SERVER] ||
      !self->poll)
    {
      z_proxy_quit(&self->super);
      return;
    }
    
  z_fd_set_nonblock(self->super.endpoints[EP_CLIENT]->fd, 1);
  self->super.endpoints[EP_CLIENT]->read = plug_copy_client_to_server;
  self->super.endpoints[EP_CLIENT]->write = plug_copy_server_to_client;
  self->super.endpoints[EP_CLIENT]->user_data = self;
  self->super.endpoints[EP_CLIENT]->want_read = TRUE;
  self->super.endpoints[EP_CLIENT]->timeout = -2;
  
  
  z_fd_set_nonblock(self->super.endpoints[EP_SERVER]->fd, 1);
  self->super.endpoints[EP_SERVER]->read = plug_copy_server_to_client;
  self->super.endpoints[EP_SERVER]->write = plug_copy_client_to_server;
  self->super.endpoints[EP_SERVER]->user_data = self;
  self->super.endpoints[EP_SERVER]->want_read = TRUE;
  self->super.endpoints[EP_SERVER]->timeout = -2;

  z_poll_add_stream(self->poll, self->super.endpoints[EP_CLIENT]);
  z_poll_add_stream(self->poll, self->super.endpoints[EP_SERVER]);
  if (self->downstreams[EP_CLIENT])
    {
      self->downbufs[EP_CLIENT] = g_new0(ZIOBuffer, 1);
      self->downbufs[EP_SERVER] = g_new0(ZIOBuffer, 1);

      self->super.endpoints[EP_CLIENT]->read = plug_copy_client_to_down;
      self->super.endpoints[EP_CLIENT]->write = plug_copy_down_to_client;

      self->super.endpoints[EP_SERVER]->read = plug_copy_server_to_down;
      self->super.endpoints[EP_SERVER]->write = plug_copy_down_to_server;

      self->downstreams[EP_CLIENT]->read = plug_copy_down_to_client;
      self->downstreams[EP_CLIENT]->write = plug_copy_client_to_down;
      self->downstreams[EP_CLIENT]->user_data = self;
      self->downstreams[EP_CLIENT]->want_read = TRUE;

      self->downstreams[EP_SERVER]->read = plug_copy_down_to_server;
      self->downstreams[EP_SERVER]->write = plug_copy_server_to_down;
      self->downstreams[EP_SERVER]->user_data = self;
      self->downstreams[EP_SERVER]->want_read = TRUE;

      z_poll_add_stream(self->poll, self->downstreams[EP_CLIENT]);
      z_poll_add_stream(self->poll, self->downstreams[EP_SERVER]);
    }
}

/* FIXME: clean this up */
static gboolean
plug_request_stack_event(PlugProxy *self)
{
  ZPolicyObj *res;
  gboolean called;

  z_policy_lock(self->super.thread);
  res = z_policy_call(self->super.handler, "requestStack", NULL, &called);
  if (res)
    {
      /* requestStack should return a reference to a function requiring two
       * parameters: client and server fd */
      
      if (PyCallable_Check(res))
        {
          int downpair[2], uppair[2];
          
          if (socketpair(AF_UNIX, SOCK_STREAM, 0, downpair) == -1)
            {
              z_message(0, "%s: Error creating socketpair: %m", self->super.session_id);
              return FALSE;
            }
          else if (socketpair(AF_UNIX, SOCK_STREAM, 0, uppair) == -1)
            {
              close(downpair[0]);
              close(downpair[1]);
              z_message(0, "%s: Error creating socketpair: %m", self->super.session_id);
              return FALSE;
            }
          else
            {
              ZPolicyObj *res2;
              
              res2 = PyEval_CallFunction(res, "(ii)", downpair[1], uppair[1]);
              if (!res2)
                {
                  if (PyErr_Occurred())
                    PyErr_Print();
                  close(downpair[0]);
                  close(downpair[1]);
                  close(uppair[0]);
                  close(uppair[1]);
                  return FALSE;
                }
              else
                {
                  z_fd_set_nonblock(downpair[0], 1);
                  z_fd_set_nonblock(uppair[0], 1);
                  self->downstreams[EP_CLIENT] = z_stream_new(downpair[0], "plug client downstream");
                  self->downstreams[EP_SERVER] = z_stream_new(uppair[0], "plug server downstream");
                }
              z_policy_var_unref(res2);
                
            }
        }
      else
        {
          guint rv;
          if(z_policy_var_parse(res,"i",&rv))
            {
              if(rv!=0)
                {
                  z_proxy_log(self, "plug.error", 1, "Internal error. Bad return from requestStack");
                  return FALSE;
                }
            }
          else
            {
              z_proxy_log(self,"plug.error", 1, "Internal error. Bad return type from requestStack");
              return FALSE;
            }
        }
    }
  else
    if (called)
      return FALSE;
  z_policy_var_unref(res);
  z_policy_unlock(self->super.thread);
  return TRUE;
}

static void
plug_thread(gpointer s)
{
  PlugProxy *self = (PlugProxy *) s;
  
  gettimeofday(&self->started_time, NULL);

  ZPROXY_SET_STATE(self, PS_CONFIG);
  z_proxy_config_event(&self->super);

  ZPROXY_SET_STATE(self, PS_STARTING_UP);
  z_proxy_startup_event(&self->super);

  /* this sets the server side endpoint if successful */
  z_proxy_connect_server_event(&self->super, NULL, 0);
  
  if (!plug_request_stack_event(self))
    ZPROXY_SET_FLAG(self, PF_QUIT);

  else
    {
      ZPROXY_SET_STATE(self, PS_WORKING);

      self->poll = z_poll_new();
      plug_init_streams(self);
  
      while (!(self->super.flags & PF_QUIT) && z_poll_iter_timeout(self->poll, self->timeout))
        {
          ;
        }
    }
 
  ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event(&self->super);
  
  z_poll_unref(self->poll);
  self->poll = NULL;

  z_proxy_destroy(&self->super);
}


ZProxy *
plug_proxy_new(gchar *session_id,
	       ZStream *client,
	       ZPolicyObj *handler)
{
  PlugProxy *self = g_new0(PlugProxy, 1);

  z_proxy_init(&self->super, session_id, client, handler);
  
  z_proxy_var_new(&self->super, "timeout", Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT, &self->timeout);
  z_proxy_var_new(&self->super, "copy_to_client", Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT, &self->copy_to_client);
  z_proxy_var_new(&self->super, "copy_to_server", Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT, &self->copy_to_server);
  z_proxy_var_new(&self->super, "packet_stats_interval", Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT, &self->packet_stats_interval);

  self->super.free = plug_proxy_free;
  self->copy_to_server = 1;
  self->copy_to_client = 1;
  self->timeout = -1;
  z_thread_new(session_id, plug_thread, self);
  return &self->super;
}

static void
plug_proxy_free(ZProxy *s)
{
  PlugProxy *self = (PlugProxy *) s;
  
  int i;
  for (i = 0; i < EP_MAX; i++)
    {
      if (self->downstreams[i])
        {
          z_stream_shutdown(self->downstreams[i], SHUT_RDWR);
          z_stream_close(self->downstreams[i]);
          z_stream_unref(self->downstreams[i]);
          g_free(self->downbufs[i]);
        }
    }
}

gint
zorp_module_init(void)
{
  z_registry_add("plug", ZR_PROXY, plug_proxy_new);
  return TRUE;
}
