/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pssl.c,v 1.9.2.3 2001/04/06 14:55:47 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/stream.h>
#include <zorp/proxy.h>
#include <zorp/poll.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/log.h>
#include <zorp/registry.h>
#include <zorp/sockaddr.h>
#include <zorp/io.h>
#include <zorp/giossl.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <openssl/err.h>

#define ENABLE_PKT_STATS 1

#define PROXY_BUFSIZE    1500	/* 64k */

#define SSL_ERROR "ssl.error"

#define EOF_CLIENT         0x0001
#define EOF_SERVER         0x0002
#define EOF_STACKED_CLIENT 0x0004
#define EOF_STACKED_SERVER 0x0008

typedef struct _ZIOBuffer
{
  gchar buf[PROXY_BUFSIZE];
  guint ofs, end;
  guint packet_count, packet_bytes;
} ZIOBuffer;

typedef struct _PsslProxy
{
  ZProxy super;
  ZStream *downstreams[EP_MAX];
  ZIOBuffer buffers[EP_MAX];
  ZIOBuffer *downbufs[EP_MAX];
  ZSockAddr *addr;
  ZPoll *poll;
  gint eofmask;
  gint timeout;
  gint packet_stats_interval;
  guint global_packet_count;
  gboolean copy_to_server, copy_to_client;
  GString *client_keyfile, *client_certfile;
  GString *server_keyfile, *server_certfile;
  gboolean need_ssl[EP_MAX];
  struct timeval started_time;
  SSL_CTX *client_ctx, *server_ctx;
  X509* peer_cert;
} PsslProxy;

typedef ZStream ZRealStream;

static void pssl_proxy_free(ZProxy * s);

#define PSSL_DEBUG "pssl.debug"
#define PSSL_ERROR "pssl.error"

static guint
pssl_read_input(PsslProxy *self, ZStream *input, ZIOBuffer *buf)
{
  gint rc;
  gboolean called;

  rc = z_stream_read(input, buf->buf, sizeof(buf->buf), &buf->end);
  switch (rc)
    {
    case G_IO_ERROR_NONE:
      buf->packet_bytes += buf->end;
      buf->packet_count++;
      self->global_packet_count++;
      if (self->packet_stats_interval &&
	  (self->global_packet_count % self->packet_stats_interval) == 0)
	{
	  int i;
	  ZPolicyObj *res;

	  z_policy_lock(self->super.thread);
	  for (i = EP_CLIENT; i < EP_MAX; i++)
	    {
	      ZIOBuffer *buf;
	      struct timeval now, spent;
	      char *varname[EP_MAX] =
		{ "bandwidth_to_client", "bandwidth_to_server" };

	      buf = &self->buffers[i];
	      gettimeofday(&now, NULL);
	      spent.tv_sec = now.tv_sec - self->started_time.tv_sec;
	      spent.tv_usec = now.tv_usec - self->started_time.tv_usec;
	      if (spent.tv_usec < -500000)
		{
		  spent.tv_sec++;
		}
	      z_policy_setattr(self->super.handler, varname[i],
			       z_policy_var_build("d",
						  ((double) buf->
						   packet_bytes) /
						  spent.tv_sec));
	    }
	  res =
	    z_policy_call(self->super.handler, "packetStats",
			   z_policy_var_build("iiii",
					      self->buffers[EP_CLIENT].
					      packet_bytes,
					      self->buffers[EP_CLIENT].
					      packet_count,
					      self->buffers[EP_SERVER].
					      packet_bytes,
					      self->buffers[EP_SERVER].
					      packet_count),
			   &called);
	  if (res)
	    {
              if (!z_policy_var_parse(res, "i", &rc))
                rc = Z_DENY;
            }
          else
            if (called)
              rc = Z_DENY;
          z_policy_var_unref(res);
	  z_policy_unlock(self->super.thread);
	  if (rc != Z_ACCEPT)
	    {
	      return ST_CLOSE | ST_EOF;
	    }
	}
      if (buf->end == 0)
	{
	  return ST_CLOSE | ST_EOF;
	}
      return ST_OK | ST_GOON;
    case G_IO_ERROR_AGAIN:
      return ST_AGAIN | ST_GOON;
    default:
      return ST_ERR | ST_CLOSE;
    }
}

static guint
pssl_write_output(PsslProxy *self, ZIOBuffer *buf, ZStream *output)
{
  int rc;
  guint bytes_written;

  if (buf->ofs != buf->end)
    {
      /* buffer not empty */
      rc =
	z_stream_write(output, &buf->buf[buf->ofs], buf->end - buf->ofs,
		       &bytes_written);
      switch (rc)
	{
	case G_IO_ERROR_NONE:
	  buf->ofs += bytes_written;
	  break;
	case G_IO_ERROR_AGAIN:
	  break;
	default:
	  return ST_ERR | ST_CLOSE;
	}
      if (buf->ofs != buf->end)
	{
	  output->want_write = TRUE;
	  return ST_AGAIN | ST_GOON;
	}
      else
	return ST_OK | ST_GOON;
    }
  return ST_OK | ST_GOON;
}

static guint
pssl_copy_data(PsslProxy *self, ZStream *from, ZStream *to,
	       ZIOBuffer *buf)
{
  int rc;
  int pkt_count = 0;

  from->want_read = FALSE;
  if (to)
    {
      to->want_write = FALSE;
      rc = pssl_write_output(self, buf, to);
      if (ST_ERROR(rc) != ST_OK)
	return rc;
    }
  while (1)
    {
      buf->ofs = buf->end = 0;
      rc = pssl_read_input(self, from, buf);
      if (ST_ERROR(rc) == ST_OK)
	{
	  if (to)
	    {
	      rc = pssl_write_output(self, buf, to);
	      if (ST_ERROR(rc) == ST_AGAIN)
		break;
	      else if (ST_FAILED(rc))
		{
		  z_proxy_log(self, PSSL_ERROR, 1, "Error writing %s (%m)",
			    to->name);
		  return ST_QUIT | ST_OK;
		}
	    }
	}
      else if (ST_ERROR(rc) == ST_AGAIN)
	break;
      else if (ST_ERROR(rc) == ST_EOF)
	return ST_EOF | ST_OK;
      else if (ST_FAILED(rc))
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Error reading %s (%m)",
		    from->name);
	  return ST_QUIT | ST_OK;
	}
      pkt_count++;
    }
  if (buf->ofs == buf->end)
    {
      from->want_read = TRUE;
    }
  return ST_OK | ST_GOON;
}

/* callbacks when no stacking is made */
static guint
pssl_copy_client_to_server(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  if (self->copy_to_server)
    ret = pssl_copy_data(self, self->super.endpoints[EP_CLIENT],
			 self->super.endpoints[EP_SERVER],
			 &self->buffers[EP_SERVER]);
  else
    ret = pssl_copy_data(self, self->super.endpoints[EP_CLIENT], NULL,
			 &self->buffers[EP_SERVER]);
  if (ST_ERROR(ret) == ST_EOF)
    {
      ret = ST_QUIT | ST_OK;
    }
  return ret;
}

static guint
pssl_copy_server_to_client(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  if (self->copy_to_client)
    ret = pssl_copy_data(self, self->super.endpoints[EP_SERVER],
			 self->super.endpoints[EP_CLIENT],
			 &self->buffers[EP_CLIENT]);
  else
    ret = pssl_copy_data(self, self->super.endpoints[EP_SERVER], NULL,
			 &self->buffers[EP_CLIENT]);
  if (ST_ERROR(ret) == ST_EOF)
    {
      ret = ST_QUIT | ST_OK;
    }
  return ret;
}

/* callbacks when a stacked module exists */
static guint
pssl_copy_client_to_down(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  if (self->copy_to_server)
    ret = pssl_copy_data(self, self->super.endpoints[EP_CLIENT],
			 self->downstreams[EP_CLIENT],
			 self->downbufs[EP_CLIENT]);
  else
    ret = pssl_copy_data(self, self->super.endpoints[EP_CLIENT], NULL,
			 self->downbufs[EP_CLIENT]);
  if (ST_ERROR(ret) == ST_EOF)
    {
      z_stream_shutdown(self->super.endpoints[EP_CLIENT], SHUT_RD);
      z_stream_shutdown(self->downstreams[EP_CLIENT], SHUT_WR);
      self->eofmask |= EOF_CLIENT;
      ret = ST_GOON | ST_OK;
    }
  return ret;
}

static guint
pssl_copy_down_to_client(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  ret = pssl_copy_data(self, self->downstreams[EP_CLIENT],
		       self->super.endpoints[EP_CLIENT],
		       &self->buffers[EP_CLIENT]);

  if (ST_ERROR(ret) == ST_EOF)
    {
      z_stream_shutdown(self->downstreams[EP_CLIENT], SHUT_RD);
      z_stream_shutdown(self->super.endpoints[EP_CLIENT], SHUT_WR);
      self->eofmask |= EOF_STACKED_CLIENT;
      if ((self->eofmask & (EOF_STACKED_CLIENT + EOF_STACKED_SERVER)) == (EOF_STACKED_CLIENT + EOF_STACKED_SERVER))
        ret = ST_QUIT | ST_OK;
      else
        ret = ST_GOON | ST_OK;
    }
  return ret;
}

static guint
pssl_copy_server_to_down(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  if (self->copy_to_client)
    ret = pssl_copy_data(self, self->super.endpoints[EP_SERVER],
			 self->downstreams[EP_SERVER],
			 self->downbufs[EP_SERVER]);
  else
    ret = pssl_copy_data(self, self->super.endpoints[EP_SERVER], NULL,
			 self->downbufs[EP_SERVER]);

  if (ST_ERROR(ret) == ST_EOF)
    {
      z_stream_shutdown(self->super.endpoints[EP_SERVER], SHUT_RD);
      z_stream_shutdown(self->downstreams[EP_SERVER], SHUT_WR);
      self->eofmask |= EOF_SERVER;
      ret = ST_GOON | ST_OK;
    }
  return ret;
}

static guint
pssl_copy_down_to_server(ZStream *stream, GIOCondition cond)
{
  PsslProxy *self = (PsslProxy *) stream->user_data;
  guint ret;

  ret = pssl_copy_data(self, self->downstreams[EP_SERVER],
		       self->super.endpoints[EP_SERVER],
		       &self->buffers[EP_SERVER]);
  if (ST_ERROR(ret) == ST_EOF)
    {
      z_stream_shutdown(self->downstreams[EP_SERVER], SHUT_RD);
      z_stream_shutdown(self->super.endpoints[EP_SERVER], SHUT_WR);
      self->eofmask |= EOF_STACKED_SERVER;
      if ((self->eofmask & (EOF_STACKED_CLIENT + EOF_STACKED_SERVER)) == (EOF_STACKED_CLIENT + EOF_STACKED_SERVER))
        ret = ST_QUIT | ST_OK;
      else
        ret = ST_GOON | ST_OK;
    }
  return ret;
}

static void
pssl_init_streams(PsslProxy *self)
{
  SSL *ssl;
  int ret;
  
  if (!self->super.endpoints[EP_CLIENT] || 
      !self->super.endpoints[EP_SERVER] ||
      !self->poll)
    {
      z_proxy_quit(&self->super);
      return;
    }

  if (self->need_ssl[EP_CLIENT])
    {
      ssl = NULL;
      z_proxy_log(self, PSSL_DEBUG, 6, "Client needs ssl.");
      self->server_ctx = SSL_CTX_new(SSLv23_server_method());
      if (!self->server_ctx)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Error loading SSL methods.");
	  z_proxy_quit(&self->super);
	  return;
	}
      SSL_CTX_set_options(self->server_ctx,SSL_MODE_ENABLE_PARTIAL_WRITE);
      ret =
	SSL_CTX_use_PrivateKey_file(self->server_ctx, self->client_keyfile->str,
				    SSL_FILETYPE_PEM);
      if (ret < 0)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Error loading keyfile: %s.", self->client_keyfile->str);
	  z_proxy_quit(&self->super);
	  return;
	}
      z_proxy_log(self, PSSL_DEBUG, 6, "Client private key %s loaded.", self->client_keyfile->str);
      ret =
	SSL_CTX_use_certificate_file(self->server_ctx, self->client_certfile->str,
				     SSL_FILETYPE_PEM);
      if (ret < 0)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Error loading certificate (%s)",
		    self->client_certfile->str);
	  z_proxy_quit(&self->super);
	  return;
	}
      z_proxy_log(self, "PSSL_DEBUG", 6, "Client side certificate file %s loaded.", self->client_certfile->str);
      ret = SSL_CTX_check_private_key(self->server_ctx);
      if (ret < 0)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Client side certificate not signed with private key.");
	  z_proxy_quit(&self->super);
	  return;
	}
      z_proxy_log(self, PSSL_DEBUG, 6, "Private key check passed");
      ssl = SSL_new(self->server_ctx);
      if (!ssl)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Not enough memory for SSL, or bad CTX.");
	  z_proxy_quit(&self->super);
	  return;
	}
      SSL_set_fd(ssl, self->super.endpoints[EP_CLIENT]->fd);
      ret = SSL_accept(ssl);
      if (ret < 0)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Client side SSL handshake failed.");
	  z_proxy_quit(&self->super);
	  return;
	}
      z_proxy_log(self, PSSL_DEBUG, 6, "Client side SSL handshake successful.");
      z_stream_ssl(self->super.endpoints[EP_CLIENT], ssl);
    }
  if (self->need_ssl[EP_SERVER])
    {
      ssl = NULL;
      
      z_proxy_log(self, PSSL_DEBUG, 6, "Server needs ssl.");
      self->client_ctx = SSL_CTX_new(SSLv23_client_method());
      if (!self->client_ctx)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Error loading SSL methods.");
	  z_proxy_quit(&self->super);
	  return;
	}
      SSL_CTX_set_options(self->client_ctx,SSL_MODE_ENABLE_PARTIAL_WRITE);
      if (self->server_keyfile->len)
        {
	  ret =
	    SSL_CTX_use_PrivateKey_file(self->client_ctx, self->server_keyfile->str,
					SSL_FILETYPE_PEM);
	  if (ret < 0)
	    {
	      z_proxy_log(self, PSSL_ERROR, 1, "Error loading keyfile: %s.", self->server_keyfile->str);
	      z_proxy_quit(&self->super);
	      return;
	    }
	  z_proxy_log(self, PSSL_DEBUG, 6, "Server side private key %s loaded.", self->server_keyfile->str);
	  
	  ret =
	    SSL_CTX_use_certificate_file(self->client_ctx, self->server_certfile->str,
					 SSL_FILETYPE_PEM);
	  if (ret < 0)
	    {
	      z_proxy_log(self, PSSL_ERROR, 1, "Error loading certificate: %s", self->server_certfile->str);
	      z_proxy_quit(&self->super);
	      return;
	    }
	  z_proxy_log(self, PSSL_DEBUG, 6, "Server side certificate file %s loaded", self->server_certfile->str);
	  ret = SSL_CTX_check_private_key(self->client_ctx);
	  if (ret < 0)
	    {
	      z_proxy_log(self, PSSL_ERROR, 1, "Server side certificate is not signed with private key.");
	      z_proxy_quit(&self->super);
	      return;
	    }
	  z_proxy_log(self, PSSL_DEBUG, 6, "Server side private key check passed");
	}
      ssl = SSL_new(self->client_ctx);
      if (!ssl)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Not enough memory for SSL, or bad CTX.");
	  z_proxy_quit(&self->super);
	  return;
	}
      SSL_set_fd(ssl, self->super.endpoints[EP_SERVER]->fd);
      ret = SSL_connect(ssl);
      if (ret < 0)
	{
	  z_proxy_log(self, PSSL_ERROR, 1, "Server side SSL handshake failed.");
	  z_proxy_quit(&self->super);
	  return;
	}
      z_proxy_log(self, PSSL_DEBUG, 6, "Server side SSL handshake successful.");
      z_stream_ssl(self->super.endpoints[EP_SERVER], ssl);
    }

  z_fd_set_nonblock(self->super.endpoints[EP_CLIENT]->fd, 1);
  self->super.endpoints[EP_CLIENT]->read = pssl_copy_client_to_server;
  self->super.endpoints[EP_CLIENT]->write = pssl_copy_server_to_client;
  self->super.endpoints[EP_CLIENT]->user_data = self;
  self->super.endpoints[EP_CLIENT]->want_read = TRUE;

  z_fd_set_nonblock(self->super.endpoints[EP_SERVER]->fd, 1);
  self->super.endpoints[EP_SERVER]->read = pssl_copy_server_to_client;
  self->super.endpoints[EP_SERVER]->write = pssl_copy_client_to_server;
  self->super.endpoints[EP_SERVER]->user_data = self;
  self->super.endpoints[EP_SERVER]->want_read = TRUE;

  z_poll_add_stream(self->poll, self->super.endpoints[EP_CLIENT]);
  z_poll_add_stream(self->poll, self->super.endpoints[EP_SERVER]);
  if (self->downstreams[EP_CLIENT])
    {
      self->downbufs[EP_CLIENT] = g_new0(ZIOBuffer, 1);
      self->downbufs[EP_SERVER] = g_new0(ZIOBuffer, 1);

      self->super.endpoints[EP_CLIENT]->read = pssl_copy_client_to_down;
      self->super.endpoints[EP_CLIENT]->write = pssl_copy_down_to_client;

      self->super.endpoints[EP_SERVER]->read = pssl_copy_server_to_down;
      self->super.endpoints[EP_SERVER]->write = pssl_copy_down_to_server;

      self->downstreams[EP_CLIENT]->read = pssl_copy_down_to_client;
      self->downstreams[EP_CLIENT]->write = pssl_copy_client_to_down;
      self->downstreams[EP_CLIENT]->user_data = self;
      self->downstreams[EP_CLIENT]->want_read = TRUE;

      self->downstreams[EP_SERVER]->read = pssl_copy_down_to_server;
      self->downstreams[EP_SERVER]->write = pssl_copy_server_to_down;
      self->downstreams[EP_SERVER]->user_data = self;
      self->downstreams[EP_SERVER]->want_read = TRUE;

      z_poll_add_stream(self->poll, self->downstreams[EP_CLIENT]);
      z_poll_add_stream(self->poll, self->downstreams[EP_SERVER]);
    }
}

static gboolean
pssl_request_stack_event(PsslProxy * self)
{
  ZPolicyObj *res;
  gboolean called;

  z_policy_lock(self->super.thread);
  res = z_policy_call(self->super.handler, "requestStack", NULL, &called);
  if (res)
    {
      /* requestStack should return a reference to a function requiring two
       * parameters: client and server fd */

      if (PyCallable_Check(res))
	{
	  int downpair[2], uppair[2];

	  if (socketpair(AF_UNIX, SOCK_STREAM, 0, downpair) == -1)
	    {
	      z_message(0, "%s: Error creating socketpair: %m",
			self->super.session_id);
	    }
	  else if (socketpair(AF_UNIX, SOCK_STREAM, 0, uppair) == -1)
	    {
	      close(downpair[0]);
	      close(downpair[1]);
	      z_message(0, "%s: Error creating socketpair: %m",
			self->super.session_id);
	    }
	  else
	    {
	      ZPolicyObj *res2;

	      res2 = PyEval_CallFunction(res, "(ii)", downpair[1], uppair[1]);
	      if (!res2)
		{
		  if (PyErr_Occurred())
		    PyErr_Print();
		  close(downpair[0]);
		  close(downpair[1]);
		  close(uppair[0]);
		  close(uppair[1]);
		}
	      else
		{
		  z_fd_set_nonblock(downpair[0], 1);
		  z_fd_set_nonblock(uppair[0], 1);
		  self->downstreams[EP_CLIENT] =
		    z_stream_new(downpair[0], "pssl client downstream");
		  self->downstreams[EP_SERVER] =
		    z_stream_new(uppair[0], "pssl server downstream");
		}
	      z_policy_var_unref(res2);

	    }
	}
    }
  else
    if (called)
      return FALSE;
  z_policy_var_unref(res);
  z_policy_unlock(self->super.thread);
  return TRUE;
}

static void
pssl_thread(gpointer s)
{
  PsslProxy *self = (PsslProxy *) s;

  gettimeofday(&self->started_time, NULL);

  ZPROXY_SET_STATE(self, PS_CONFIG);
  if(!z_proxy_config_event(&self->super))
    {
      z_proxy_log(self, SSL_ERROR, 1, "config event error");
      ZPROXY_SET_FLAG(self,PF_QUIT);
    }
  else
    {

      ZPROXY_SET_STATE(self, PS_STARTING_UP);
      if(!z_proxy_startup_event(&self->super))
        {
          z_proxy_log(self, SSL_ERROR, 1, "startup event error");
          ZPROXY_SET_FLAG(self,PF_QUIT);
        }
      else
        {

          /* this sets the server side endpoint if successful */
          if(!z_proxy_connect_server_event(&self->super, NULL, 0))
            {
              z_proxy_log(self, SSL_ERROR, 1, "cannot conect to server");
              ZPROXY_SET_FLAG(self,PF_QUIT);
            }
          else
            {

              if(!pssl_request_stack_event(self))
                {
                  z_proxy_log(self, SSL_ERROR, 1, "cannot stack needed proxy");
                  ZPROXY_SET_FLAG(self,PF_QUIT);
                }

              else
                {
                  ZPROXY_SET_STATE(self, PS_WORKING);

                  self->poll = z_poll_new();
                  pssl_init_streams(self);
                  
                }
            }
        }
    }

  while (!(ZPROXY_GET_FLAG(self,PF_QUIT)) && z_poll_iter_timeout(self->poll, self->timeout))
    {
      ;
    }

  ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event(&self->super);

  z_poll_unref(self->poll);
  self->poll = NULL;

  z_proxy_destroy(&self->super);
}


static ZProxy *
pssl_proxy_new(gchar * session_id, ZStream * client, ZPolicyObj * handler)
{
  int i;
  PsslProxy *self = g_new0(PsslProxy, 1);

  z_proxy_init(&self->super, session_id, client, handler);

  self->client_keyfile = g_string_new("");
  self->client_certfile = g_string_new("");

  self->server_keyfile = g_string_new("");
  self->server_certfile = g_string_new("");
  self->timeout = -1;

  z_proxy_var_new(&self->super, "copy_to_client",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->copy_to_client);
  z_proxy_var_new(&self->super, "copy_to_server",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->copy_to_server);
  z_proxy_var_new(&self->super, "packet_stats_interval",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->packet_stats_interval);

  z_proxy_var_new(&self->super, "timeout",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->timeout);
		  
  z_proxy_var_new(&self->super, "server_need_ssl",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->need_ssl[EP_SERVER]);
  z_proxy_var_new(&self->super, "client_need_ssl",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
		  &self->need_ssl[EP_CLIENT]);
		  
  z_proxy_var_new(&self->super, "client_key",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING,
		  self->client_keyfile);
  z_proxy_var_new(&self->super, "client_cert",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING,
		  self->client_certfile);

  z_proxy_var_new(&self->super, "server_key",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING,
		  self->client_keyfile);
  z_proxy_var_new(&self->super, "server_cert",
		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING,
		  self->client_certfile);
		  
  self->super.free = pssl_proxy_free;
  self->copy_to_server = 1;
  self->copy_to_client = 1;
  self->client_ctx = NULL;
  self->server_ctx = NULL;
  for (i = 0; i < EP_MAX; i++)
    {
      self->need_ssl[i] = 0;
    }
  z_thread_new(session_id, pssl_thread, self);
  return &self->super;
}

static void
pssl_proxy_free(ZProxy * s)
{
  PsslProxy *self = (PsslProxy *) s;

  int i;
  for (i = 0; i < EP_MAX; i++)
    {
      if (self->downstreams[i])
	{
	  z_stream_shutdown(self->downstreams[i], SHUT_RDWR);
	  z_stream_close(self->downstreams[i]);
	  z_stream_unref(self->downstreams[i]);
	  g_free(self->downbufs[i]);
	}
    }
  g_string_free(self->client_keyfile, TRUE);
  g_string_free(self->client_certfile, TRUE);
  g_string_free(self->server_keyfile, TRUE);
  g_string_free(self->server_certfile, TRUE);

  if (self->server_ctx)
    {
      SSL_CTX_free(self->server_ctx);
    }
  if (self->client_ctx)
    {
      SSL_CTX_free(self->client_ctx);
    }
  ERR_remove_state(0);
}

gint
zorp_module_init(void)
{
  z_registry_add("pssl", ZR_PROXY, pssl_proxy_new);
  return TRUE;
}
