/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: sink.c,v 1.10 2001/02/19 14:28:41 bazsi Exp $
 *
 ***************************************************************************/

/* This is an experimental UDP relay (transparent proxy) */

#include <zorp/zorp.h>
#include <zorp/log.h>
#include <zorp/proxy.h>
#include <zorp/thread.h>
#include <zorp/policy.h>
#include <zorp/packsock.h>
#include <zorp/pysockaddr.h>
#include <zorp/registry.h>

typedef struct _ZSinkProxy
{
  ZProxy super;
  ZPolicyObj *client, *server;
} ZSinkProxy;

static void
sink_thread(gpointer s)
{
  char buf[128], buf2[128];
  ZSinkProxy *self = (ZSinkProxy *) s;
  PyObject *py_from = NULL, *py_to = NULL;
  ZPacket packet;
  gint read_length;
  gint locked;

  ZPROXY_SET_STATE(self, PS_CONFIG);
  z_proxy_config_event(&self->super);

  ZPROXY_SET_STATE(self, PS_STARTING_UP);
  z_proxy_startup_event(&self->super);

  z_debug(10, "sink input fd: %d", self->super.endpoints[EP_CLIENT]->fd);
  while (1)
    {
      int res;

      locked = 0;
      res = z_stream_read(self->super.endpoints[EP_CLIENT], (void *) &packet,
			  sizeof(packet), &read_length);
      if (res == G_IO_ERROR_AGAIN)
	continue;
      else if (res != G_IO_ERROR_NONE)
	{
	  z_warning(0, "Error reading UDP source: %m");
	  break;
	}
      z_message(1, "%s: input packet from=%s, to=%s, length=%d",
		self->super.session_id,
		z_sockaddr_format(packet.from, buf, 128),
		z_sockaddr_format(packet.to, buf2, 128), packet.length);

      if (read_length != sizeof(packet))
	{
	  z_warning(0, "%s: internal error, unexpected ZPacket length = %d",
		    self->super.session_id, read_length);
	  goto free_packet;
	}

      z_policy_lock(self->super.thread);
      locked = 1;
      if (!(py_from = z_zorp_sockaddr_new(packet.from)))
	{
	  z_warning(0,
		    "%s: internal error, error converting sockaddr to python\n",
		    self->super.session_id);
	  goto free_packet;
	}
      if (!(py_to = z_zorp_sockaddr_new(packet.to)))
	{
	  z_warning(0,
		    "%s: internal error, error converting sockaddr to python\n",
		    self->super.session_id);
	  goto free_packet;
	}

      res = z_policy_event(self->super.handler, "packet",
			   z_policy_var_build("(OOi)",
					      py_from, py_to, packet.length));

      if (res == Z_ACCEPT || res == Z_UNSPEC)
	{
	  z_message(1, "%s: output packet from=%s, to=%s, length=%d",
		    self->super.session_id,
		    z_sockaddr_format(packet.from, buf, 128),
		    z_sockaddr_format(packet.to, buf2, 128), packet.length);
	  if (z_packsock_sendfrom(packet.from,
			      packet.to, packet.data, packet.length) < 0)
            {
              z_warning(1, "%s: z_packsock_sendfrom failed, error=%m", self->super.session_id);
            }
	}

    free_packet:
      if (locked)
	z_policy_unlock(self->super.thread);
      z_policy_var_unref(py_from);
      z_policy_var_unref(py_to);
      z_sockaddr_unref(packet.from);
      z_sockaddr_unref(packet.to);
      g_free(packet.data);
    }

  ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
  z_proxy_shutdown_event(&self->super);

  z_proxy_destroy(&self->super);
}

ZProxy *
sink_proxy_new(gchar * session_id, ZStream * client, ZPolicyObj * handler)
{
  ZSinkProxy *self = g_new0(ZSinkProxy, 1);

  z_proxy_init(&self->super, session_id, client, handler);
  z_thread_new(session_id, sink_thread, self);
  return &self->super;
}

gint
zorp_module_init(void)
{
  z_registry_add("sink", ZR_PROXY, sink_proxy_new);
  return TRUE;
}
