############################################################################
##
## Copyright (c) 2000-2001 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Chainer.py,v 1.38.2.4 2001/03/09 14:17:12 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.18
## Notes:
##   - comments in English
##
############################################################################

"""Module implementing different methods for connecting to the server end of a connection.

This module defines an abstract interface (class 'AbstractChainer') used by
proxies to connect to their server endpoint, and some derived classes
which fill the abstract class with some real function, most notably
TransparentChainer and DirectedChainer.
"""

from Zorp import *
from Stream import Stream
from SockAddr import SockAddrInet
from Connector import Connect
import socket

class AbstractChainer:
	"""Class encapsulating the establishment of server-side connections.
	
	Proxies are organized hierarchically, where the top level proxy is
	directly connected to server, and the low level proxy is connected
	to the upper level proxy. The term chaining means establishing 
	the connection between a proxy and its parent. A chainer is 
	a class responsible for chaining.
	
	A chainer is only required for top level proxies, but may present in
	stacked ones too. The default chaining behaviour when there's no
	chainer defined (ie.  'self.session.chainer' is 'None') is to pick
	up 'self.session.server_fd', which in turn gets set when the stacked
	proxy is started. 
	
	The task of chainers used for top level proxies is to connect to the
	target server. 
	
	Attributes
	
	  none
	
	"""
	def connectServer(self, session, host, port):
		"""Function to perform chaining to the parent proxy or remote server.
		
		This is an interface function to be called to chain up
		to the parent proxy (or to the remote server)

		The parameters 'host' and 'port' are hints given by the
		proxy, they may be ignored (like in TransparentChainer)
		or taken into account (like in InbandChainer).

		Arguments

		  self        -- this instance
		  
		  session     -- Session object
		  
		  host	      -- host to connect to (sockaddr).
		  
		  port        -- port to connect to
		  
		"""
		raise NotImplementedError
		
	def establishConnection(self, session, local, remote):
		"""Function to actually establish a connection.
		
		Internal function to establish a connection with the given
		local and remote addresses. It is used by derived chainer
		classes after finding out the destination address to connect
		to. This function performs access control checks.
		   
		Arguments

		  self         -- this instance
		  
		  session      -- Session object
		  
		  local	       -- bind address
		  
		  remote       -- host to connect to
		  
		Returns
		
		  The fd of the connection to the server

		"""
		try:
			session.setServer(remote)
			if session.isServerPermitted() == Z_ACCEPT:
				#remote.options = session.client_address.options
				conn = Connect(local, remote)
				session.server_fd = conn.blockingConnect()
				session.server_local = conn.local
				log(CORE_SESSION, 3, "%s: server_fd=%d, server_addr=%s, server_zone=%s, server_local=%s" % (session.session_id, session.server_fd, session.server_address, session.server_zone, session.server_local))
				return session.server_fd
			raise DACException
		except DACException:
			pass
		except MACException:
			pass
		return -1

		

class DirectedChainer(AbstractChainer):
	"""Class implementing connection to a predefined server.
	
	This chainer connects to a predefined server determined at instance
	creation time, independently of the original destination of the
	connection request. It can be used to address a nonrouteable host
	from the firewall.
	   
	Attributes
	
	  local	     -- The local data of the connection
	  
	  remote     -- The data of the server
	  
	  forge_addr -- forge client address when connecting to the server

	Examples

	  Example (Sample for DirectedChainer usage)

	    Service("inter_HTTP_dmz",
	            DirectedChainer(SockAddr('192.168.0.2', 80)),
		    HttpProxy)

	"""
	
	def __init__(self, remote, local = None, forge_addr = 0):
		"""Constructor to initialize a DirectedChainer instance.

		Sets instance attributes based on parameters.

		Arguments

		  self       -- this instance
		
		  local	     -- Local address of the connection
		  
		  remote     -- Address of the server
		  
		  forge_addr -- forge client address when connecting to
		                the server
		"""
		try:
			x = remote.ip
			x = remote.port
			del x
		except AttributeError:
			raise ValueError, "remote must be a SockAddrInet instance"
		self.local = local
		self.remote = remote
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""Function to connect to the predefined remote server.

		This function establishes connection to the
		remote server forging the client address
		if necessary.

		Arguments

		  self         -- this instance
		
		  session      -- The data of the current session
		  
		  host	       -- not used
		  
		  port	       -- not used
			
		Returns
		
		  The fd given by establishConnection
		"""

		if self.forge_addr:
			local = session.client_address
			local.port = 0
		else:
			local = self.local

		return self.establishConnection(session, local, self.remote)

class TransparentChainer(AbstractChainer):
	"""Class encapsulating connection to the client's original destination.

	This chainer connects to the destination originally addressed by the
	client and is used for connections intercepted by the firewall.
	
	You have to add a REDIRECT rule to your packet filter
	configuration to force connections going through your
	firewall to be serviced locally.

	Make sure that the Listener port is protected by a DENY
	rule so that clients cannot connect to it directly, otherwise
	TransparentProxy may cause an infinite loop.

	Attributes
	
	  local		      -- bind address

	  forge_addr	      -- use client address as outgoing 
	                         local address

	  forced_port	      -- if not 0 force it as destination port


	Examples

          Example (Sample for TransparentChainer usage)

	    Service("http", TransparentChainer(), HttpProxy)
	"""
	
	def __init__(self, local = None, forge_addr = 0, forced_port = 0):
		"""Constructor to initialize a TransparentChainer instance.
		   
		Sets various attributes as passed in as parameters.
		
		Arguments

		  self        -- this instance
		
		  local       -- local address, may be none (optional)
		  
		  forge_addr  -- if the client address shall be forged (optional)
		  
		  forced_port -- use this port on the remote server (optional)
		"""
		
		self.local = local
		self.forge_addr = forge_addr
		self.forced_port = forced_port
		
	def connectServer(self, session, host, port):
		"""Function to connect to the remote server.
		
		Finds out the original destination of the given connection
		and connects there.

		Arguments
		
		  self         -- this instance
		  
		  session      -- The data of the current session
		  
		  host	       -- not used
		  
		  port	       -- not used
			
		Returns
		
		  The fd given by establishConnection
		"""

		if self.forge_addr:
			local = session.client_address
			local.port = 0
		else:
			local = self.local
				
		remote = getTransparentDestination(session.client_fd)
		if self.forced_port != 0:
			remote.port = self.forced_port

		return self.establishConnection(session, local, remote)

class InbandChainer(AbstractChainer):
	"""Encapsulates a chainer for protocols where the connection information comes from the protocol logic.
	
	This chainer can be used for protocols where the target address
	is determined during the communication with the client. An
	example would be non-transparent HTTP where the target address
	is sent in the request.
	   
	Attributes
	
	  local       -- the address of the local end of the connection
	  
	  forge_addr  -- use the client address for outgoing local address
	"""
	
	def __init__(self, local = None, forge_addr = 0):
		"""Constructor to initialize an InbandChainer instance.

		Creates an instance, sets its attributes based on constructor
		arguments.
		
		Arguments

		  self       -- this instance
		  
		  local	     --	local address (optional)
		  
		  forge_addr --	forge client address (optional)
		  
		"""
		self.local = local
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""Function called by the proxy to connect to the remote end.

		This function looks up the host given as parameter in the
		DNS and starts connecting there.
		
		Arguments

		  self         -- this instance
		  
		  session      -- Session object
		  
		  host	       -- Remote hostname as determined by the protocol
		  
		  port	       -- Remote port as determinted by the protocol
		  
		Returns
		
		  The fd given by establishConnection
		"""
		
		#message(0, "InbandChainer.connectServer()")
		try:
			ip = socket.gethostbyname(host)
			if self.forge_addr:
				local = session.client_address
				local.port = 0
			else:
				local = self.local
			return self.establishConnection(session, local, SockAddrInet(ip, port))
		except socket.error:
			return -1

class FailoverChainer(AbstractChainer):
	"""Class encapsulating connection to a group of remote servers.
	
	Failover chainer tries to connect several hosts in round robin
	fashion until one of them succeeds, forming a simple failover and
	load balance solution.
		
	Attributes
	
	  hosts		-- The list of addresses to try
	  
	  local		-- the address of the local end of the connection
	  
	  forge_addr	-- forge client address
	  
	  current_host	-- The index of the actual host in the list

	"""
	#FIXME: support priorities, shorter than system default timeouts
	
	def __init__(self, hosts, local = None, forge_addr = 0):
		"""Constructor to initializes an instance of FailoverChainer

		Sets attributes based on parameters.
		
		Arguments
		
		  self                  -- this instance
		
		  hosts			-- The list of addresses to try
		  
		  local			-- address to bind to (Optional)
		  
		  forge_addr		-- forge client address (Optional)
		"""
			
		self.local = local
		self.forge_addr = forge_addr
		self.hosts = hosts
	
	def nextHost(self):
		"""Function to skip to the next host to try.
		   
		This function is called to skip to the next host to try in
		the hosts array. Increments 'current_host', zeroes
		it if overflowed.
		
		Arguments
		
		  self  -- this instance

		"""
		self.current_host = self.current_host + 1
		if self.current_host == len(self.hosts):
			self.current_host = 0

	def getHostAddress(self):
		"""Function to returns the actual host to attempt connection to.

		Returns the actual 'current_host' element of the host list
		'hosts'.

		Arguments
		
		  self -- this instance
		  
		Returns
		
		  address to connect to

		"""
		return self.hosts[self.current_host]
	
	def connectServer(self, session, host, port):
		"""Called by the proxy to connect to the remote server.
		   
		Tries to connect to each host round-robin until
		one of them succeeds.

		Arguments

		  self          -- this instance
		  
		  session	-- The data of the current session
		  
		  host		-- not used
		  
		  port		-- not used
		  
		Returns
		
		  The fd given by establishConnection
		
		"""

		if self.forge_addr != 0:
			local = session.client_address
			local.port = 0
		else:
			local = self.local
					
		first_attempt = self.current_host
		remote = self.getHostAddress()
		res = self.establishConnection(session, local, remote)
		self.nextHost()
		while res == -1 and self.current_host != first_attempt:
			remote = self.getHostAddress()
			res = self.establishConnection(session, local, remote)
			self.nextHost()
		return res
