############################################################################
##
## Copyright (c) 2000-2001 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Proxy.py,v 1.28.2.4 2001/02/22 14:06:27 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.13
## Notes:
##   - comments in English
##
############################################################################

"""Module defining classes encapsulating native proxies.

This module defines the class 'Proxy', encapsulating the ZorpProxy component
implemented by the Zorp core. Each protocol specific proxy is derived from
this class.
"""

from Zorp import *
from traceback import print_exc
from Stream import Stream
import new

def proxyLog(self, type, level, msg):
	"""Function to log a proxy message.

	This function logs a message with the 'session_id' prepended.

	Arguments

	  self  -- should be a Proxy derivative

	  type  -- log class

	  level -- message verbosity level

	  msg   -- msg to log
	  
	"""
	#FIXME: performance check. (Check level before parsing string)
	try:
		mymsg = "%s: %s" % (self.session.session_id, msg)
	except:
		mymsg = "Invalid proxy instance: " % msg
	log(type, level, mymsg)

class Proxy(ZorpProxy):
	"""Base class for the proxies.

	This class is inherited from an ExtensionClass implemented by the
	Zorp core. It encapsulates a protocol specific proxy. As an instance
	is created, the required module is loaded (identified by the
	name attribute), and a new thread is started to handle the connection.
	
	Attributes
	
	  session      -- descriptor of the connection
	  
	  name         -- the native proxy module used for this session
	"""
	
	def __init__(self, name, session):
		"""Constructor to initialize a Proxy instance.

		This constructor initializes a Proxy instance, sets
		attributes and calls the underlying C extensionclass
		to actually create a proxy instance.
		
		Arguments
		
		  name	  -- The type of the proxy
		  
		  session -- reference to the session
		"""
		self.session = session
		self.name = name
		session.setProxy(name)
		log(CORE_SESSION, 1, "%s: proxy_start, module=%s" % (session.session_id, name));
		try:
			ZorpProxy.__init__(self, name, session.session_id, session.client_stream)
		except:
			print_exc()
			session.destroy()

	def __del__(self):
		"""Destructor to deinitialize a Proxy instance.
		
		This destructor is called when this object instance is
		freed. It simply sends a message about this event to the
		log.
		
		Arguments
		
		  self -- this instance
		"""
		log(CORE_SESSION, 1, "%s: proxy_stop, module=%s" % (self.session.session_id, self.name))

	def __destroy__(self):
		"""Function called by the proxy core when the session is to be freed.
		
		This function is called when the proxy module is to be freed. It
		simply sends a message about this event to the log.
		
		Arguments
		
		  self -- this instance
		"""
		log(CORE_DEBUG, 6, "%s: proxy_destroy, module=%s" % (self.session.session_id, self.name))

	def addPolicy(self, klass):
		"""Adds a policy to the proxy.

		This function extends this class with another parent
		class at runtime. It can be used to add methods
		to this event handler class based on some
		runtime variable.
		
		Arguments

		  self        -- this instance
		  
		  klass	      -- the class implementing the policy

		"""
		self.__class__ = new.classobj(self.__class__.__name__, self.__class__.__bases__ + (klass,), self.__class__.__dict__)

	def connectServer(self, host, port):
		"""Function called to connect to the server endpoint or to the parent proxy.

		This function is called to establish connection with our
		server endpoint, which can be either a parent proxy we are
		stacked in, or a real connection to the server.

		If the chainer is defined, it is called with the given host
		and port parameters to connect to the remote server.

		If there is no chainer defined, it tries to return the
		server-side fd from the session. If there's no server_fd it
		raises an error.

		Arguments

		  self          -- this instance
		  
		  host		-- The address of the server
		  
		  port	       	-- The port of the server
		  
		Returns
		
		  The descriptor of the server stream

		"""
		#self.session.server_stream = None
		#self.session.server_fd = -1
		if self.session.chainer == None:
			
			# we have no chainer, the server side fd
			# should be available by now, used in stacked
			# proxies
			
			if self.session.server_fd == -1 or self.session.server_stream == None:
				raise InternalError, "No chainer and server_fd is -1 or server_stream is None"				
		else:
			fd = self.session.chainer.connectServer(self.session, host, port)
			if fd == -1:
				self.session.server_stream = None
			else:
				self.session.server_stream = Stream(self.session.server_fd, self.session.owner.session_id + ": server")

		return self.session.server_stream

class DatagramProxy(ZorpProxy):
	"""Base class for datagram based protocols.

	This class is similar to 'Proxy' with the exception that
	it's used for datagram based proxies, like Sink.

	Attributes

	  self          -- this instance
	  
	  session	-- the session we belong to
	"""
	
	def __init__(self, name, session):
		"""Initializes a DatagramProxy instance.

		Sets attributes based on arguments.

		Arguments

		  self          -- this instance
		  
		  name		-- The type of the proxy

		  session	-- The descriptor of the connection
		  
		"""
		self.session = session
		log(CORE_SESSION, 1, "%s: session_start, module=%s" % (session.session_id, name));
		try:
			ZorpProxy.__init__(self, name, session.session_id, session.client_stream)
		except:
			session.destroy()
