############################################################################
##
## Copyright (c) 2000-2001 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Service.py,v 1.24.2.3 2001/02/16 10:33:32 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
############################################################################

"""Module defining service related classes.

This module defines classes encapsulating service descriptions. In Zorp a
service defines how an established client request should be handled. When a
connection is accepted (usually by a listener, see the 'Listener' module),
the service bound to the given Listener is requested the create an instance
of itself (using the 'startInstance()' method, see below), which will handle
the connection and will take care about proxying traffic between the client
and the server.

Services are uniquely identified by their name, which is then used
as a reference to bind services to listeners. This name resolution
is done through the mapping named 'services'. As a new service
instance is created, it's name is registered in this hash.

The abstract interface a Service object is required to support is
defined in AbstractService, customized service classes can be derived
from this base class.
"""

from Stream import Stream
from Session import StackedSession
from Zorp import ServiceException, LimitException, log, CORE_SESSION

services = {}

class AbstractService:
	"""An abstract class defining the Service interface.
	
	This is an abstract class defining the interface Zorp uses, and as
	such it doesn't really do anything other than raising
	NotImplementedError exceptions in its methods. You should
	either derive a descendant from AbstractService, or use
	'Service' instead.

	Attributes
	
	  name		-- The name of the service
	  
	  instance_id	-- the session serviced by the service

	"""
	
	def __init__(self, name):
		"""Constructor to initialize an AbstractService or a derived class instance.

		This constructor sets attributes based on arguments, and
		registers this Service to the 'services' hash so that
		Listeners may resolve service names to service instances.
		
		Arguments

		  self  -- this instance
		  
		  name	--	The name of the service

		"""
		if services.has_key(name):
			raise ServiceException, "Duplicate service"
		services[name] = self
		self.name = name
		self.instance_id = 0
	
	def startInstance(self, session):
		"""Function to start an instance of this service.
		
		Abstract method to be implemented in derived classes.
		Should start an instance of the given service. A service
		instance takes care of the client connection, connects
		to the server and supervises the traffic going in either
		direction.
		
		Tasks of a service instance are implemented by classes
		derived from 'Proxy'.
		
		This method unconditionally raises a NotImplementedError
		exception to indicate that it must be overridden by
		descendant classes like 'Service'.
		   
		Arguments

		  self          --      this instance
		  
		  session	--	start service within this session

		"""
		raise NotImplementedError

	def stopInstance(self, session):
		"""Function called when an instance of this service is ended
		
		This function is called by Session.__del__ and indicates
		that a given session (instance) of this service is ended.
		
		Arguments
		
		  self -- this instance
		  
		  session -- ending session
		"""
		raise NotImplementedError
		
	def __str__(self):
		"""Function to represent this object as a string
		
		This function is called by the Python core when this object
		is used as-, or casted to a string. It simply returns
		the service name.
		
		Arguments
		
		  self -- this instance
		"""
		return self.name

class Service(AbstractService):
	"""Class implementing the interface defined in AbstractService.
	
	This service class uses the standard Zorp provided Proxy class
	to implement the tasks required by a Service instance.
		
	Attributes
	
	  chainer      -- The chainer class used for the service
	  
	  proxy_class  -- The proxy class used for the service
	  
	  auth	       -- Authentication handling class 
	                  (None if no authentication)
	"""

	def __init__(self, name, chainer, proxy_class, auth=None, max_instances=0):
		"""Initializes a Service instance.
		
		This function takes its arguments, and initializes
		appropriate attributes in self.
		   
		Arguments
		
		  self       -- this instance
		
		  name	     -- Name of this service (used in access control)
		  
		  chainer    -- The chainer class used for the service
		  
		  proxy_class--	The proxy class used for the service
		  
		  auth	     -- Authentication handling class (None if
		                no authentication)

		"""
		AbstractService.__init__(self, name)
		self.chainer = chainer
		self.proxy_class = proxy_class
		self.auth = None
		self.max_instances = max_instances
		self.num_instances = 0
		#self.labels = labels
		#self.auth = auth
	
		#if root == None:
		#	self.root_zone = root_zone
		#else:
		#	self.root_zone = root_zone

	def startInstance(self, session):
		"""Start a service instance.
		
		Called by the Listener to create an instance of this
		service.
		   
		Arguments

		  self          --      this instance
		  
		  session	--	The session object
			
		"""
		if self.max_instances != 0 and self.num_instances >= self.max_instances:
			raise LimitException
			
		self.num_instances = self.num_instances + 1
		session.started = 1
		
		#if session.auth == None:
		#	session.auth = self.auth		
		#if session.auth != None and not session.auth.authenticate(session):
		#	raise Zorp.AuthException

		session.name = self.name
		instance_id = self.instance_id
		
		# FIXME: handle wraparound
		self.instance_id = self.instance_id + 1
		
		session.setServiceInstance(instance_id)
		log(CORE_SESSION, 3, "%s: client_fd=%d, client_addr=%s, client_zone=%s, client_local=%s" % (session.session_id, session.client_fd, session.client_address, session.client_zone, session.client_local))
		ss = StackedSession(session, self.chainer)
		ss.client_stream = Stream(ss.client_fd, ss.session_id + ": client")
		self.proxy_class(ss)
	
	def stopInstance(self, session):
		if session.started:
			self.num_instances = self.num_instances - 1
#
# Egy instance_class es param helyett egy mappingot vesz at, ami zonatol
# fuggoen mas es mas eredmenyt ad
#
#class MultiService(AbstractService):
#	def __init__(self, name
